"""How to compute your own spectrum from fundamental constants"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/90_xphysics.ipynb.

# %% auto 0
__all__ = ['gaussian_convolve', 'ElementLines', 'get_element_spectrum', 'get_element_spectra']

# %% ../notebooks/90_xphysics.ipynb 42
import numpy as np 
import xraydb 

def gaussian_convolve(peak_energies, peak_intensities, x_keVs=None, std=0.01): 
    '''Convolves line spectrum defined by `peak_energies` and `peak_intensities` 
    
    with a Gaussian peak shape. '''
    
    if x_keVs is None: 
        x_keVs = np.linspace(0, 40, 10000)

    y_spectrum = np.zeros_like(x_keVs) 

    for peak_energy, peak_intensity in zip(peak_energies, peak_intensities): 

        y_spectrum += peak_intensity * np.exp(-(1 / std) * (x_keVs - peak_energy)**2)
        
    return x_keVs, y_spectrum 


class ElementLines(): 
    '''Computes fluorescence emission line energies and intensities for `element`.
    
    '''
    
    def __init__(self, element, excitation_energy_keV): 

        excitation_energy = 1000 * excitation_energy_keV

        lines = xraydb.xray_lines(element, excitation_energy=excitation_energy) 

        peak_names = []
        peak_labels = []
        peak_energies = [] 
        peak_intensities = []

        for name, line in lines.items(): 

            peak_names.append(name)

            # intensities (a.k.a. transition probablities) sum up to unity within each level 
            energy, intensity, initial_level, final_level = line  
            peak_energies.append(energy)
            label = f'{element}_{initial_level}{final_level}' 
            peak_labels.append(label)

            # get corresponding edge properties 
            edge = initial_level # IUPAC notation!  e.g. 'L1', not 'La'
            edge_energy, fluo_yield, jump_ratio = xraydb.xray_edge(element, edge) 
            jump_coeff = (jump_ratio - 1) / jump_ratio # see Volker 
            #print(f'{name}: {energy}; jump_coeff: {jump_coeff:.03f}; fluo_yield: {fluo_yield}')

            # multiplying edge jump coefficient, intensity and fluorescence yield... 
            peak_intensity = jump_coeff * intensity * fluo_yield 
            peak_intensities.append(peak_intensity) 
        
        # determine sorting according to peak_intensities... 
        self.peak_intensities = np.array(peak_intensities)
        indices = np.argsort(self.peak_intensities)[::-1] 

        # sort 
        self.peak_intensities = self.peak_intensities[indices]
        self.peak_energies = np.array(peak_energies)[indices] / 1000
        self.peak_names = np.array(peak_names)[indices]
        self.peak_labels = np.array(peak_labels)[indices]

        
def get_element_spectrum(element, excitation_energy_keV, normalize=True, x_keVs=None, std=0.01): 
    '''Compute simple excitation spectrum (no matrix effects). 
    
    Based on xraydb. 
    
    Returns: y_spectrum'''
    
    el = ElementLines(element, excitation_energy_keV)  
    
    x, y_spectrum = gaussian_convolve(el.peak_energies, el.peak_intensities, x_keVs=x_keVs, std=std)
    
    if normalize: 
        y_spectrum = y_spectrum / y_spectrum.max()
        
    if x_keVs is None: 
    
        return x, y_spectrum 
    
    else: 
        return y_spectrum 

    
def get_element_spectra(elements, x_keVs, excitation_energy_keV): 
    '''Compute theoretical emission spectrum for multiple elements. 
    
    Sorts elements according to largest (alpha) peak. Based on xraydb. 
    
    Returns: elements, element_spectra
    '''

    n_channels = len(x_keVs)
    n_elements = len(elements)

    element_spectra = np.zeros([n_elements, n_channels]) 

    for i, elem in enumerate(elements): 
        element_spectra[i] = get_element_spectrum(elem, excitation_energy_keV, x_keVs=x_keVs)

        # normalize
        element_spectra[i] = element_spectra[i] / element_spectra[i].max()


    # sort according to energy of largest (=alpha) peak
    alpha_idxs = np.argmax(element_spectra, axis=1)
    alpha_order = np.argsort(alpha_idxs) 

    elements = [elements[i] for i in alpha_order]
    element_spectra = element_spectra[alpha_order]
    
    return elements, element_spectra 




"""Find the Compton peak first"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/20_calibration.ipynb.

# %% auto 0
__all__ = ['RHODIUM_Ka', 'IRON_Ka', 'calibrate', 'detector_angle', 'find_instrument_peaks', 'compton_shift']

# %% ../notebooks/20_calibration.ipynb 13
import maxrf4u 
import matplotlib.pyplot as plt 
import numpy as np 
import scipy.signal as ssg 
import matplotlib.cm as cm 
import math

# %% ../notebooks/20_calibration.ipynb 14
RHODIUM_Ka = 20.210105052526263 # Rh_KL3 
IRON_Ka = 6.4032016008004 # Fe_KL3 

def calibrate(datastack_file, anode='Rh', prominence=0.1, tube_keV=40): 
    '''Automatic two step energy energy calibration. 
    
    In step 1 a preliminary calibration is done assuming that the  
    sensor peak is located at 0 keV and the Rhodium anode Ka peak 
    is next to it's high and broad Compton scattering peak in the sum spectrum. 
    
    This preliminary calibration the enables the identification of Fe_Ka peak in 
    the max spectrum and a second precise calibration. 
    
    Asks user confirmation to store energy calibration in datastack file. 
    
    Returns: x_keVs
    '''
    
    assert anode == 'Rh', NotImplementedError('Sorry, still need to implement other anode materials!') 
    
    ds = maxrf4u.DataStack(datastack_file)
    
    y_max = ds.read(ds.MAXRF_MAXSPECTRUM)
    y_sum = ds.read(ds.MAXRF_SUMSPECTRUM) 
    
    # LOCATE INSTRUMENT PEAK INDICES IN SUM SPECTRUM 
    
    left_peak_i, compton_peak_i, right_peak_i = find_instrument_peaks(y_sum, tube_keV=tube_keV) 
    
    # STEP 1: PRELIMINARY SENSOR + ANODE CALIBRATION TO LOCATE IRON Fe_Ka PEAK  
    
    left_keV = 0
    right_keV = RHODIUM_Ka 
    
    n_channels = len(y_sum) 
    x_indices = np.arange(n_channels)

    slope = (right_keV - left_keV) / (right_peak_i - left_peak_i) 
    offset = left_keV - slope * left_peak_i  
    keVs_precalib = slope * x_indices + offset 

    # STEP 2: PRECISE IRON Ka + ANODE CALIBRATION 
    
    max_peak_indices, shapes_dict = ssg.find_peaks(y_max, prominence=prominence) 
    peak_keVs_precalib = keVs_precalib[max_peak_indices]

    # now find index of peak nearest to Fe_Ka energy 
    iron_peak_i = np.argmin((peak_keVs_precalib - IRON_Ka)**2)
    
    # calibrate again now with iron Ka left hand peak 
    
    left_peak_i = max_peak_indices[iron_peak_i]
    left_keV = IRON_Ka

    slope = (right_keV - left_keV) / (right_peak_i - left_peak_i) 
    offset = left_keV - slope * left_peak_i  
    x_keVs = slope * x_indices + offset 
    
    # Calibration peak diagnostics 
    iron_xy = x_keVs[left_peak_i], y_max[left_peak_i]
    compton_xy = x_keVs[compton_peak_i], y_sum[compton_peak_i]
    anode_xy = x_keVs[right_peak_i], y_sum[right_peak_i] 
    
    theta = detector_angle(compton_xy[0], anode_xy[0])
    
    calib_peaks_x, calib_peaks_y = np.array([iron_xy, compton_xy, anode_xy]).T 
    
    # Create calibration plot 
    plt.ion()

    colors =  cm.tab10([0, 1, 2]) 
    fig, ax = plt.subplots(ncols=1, nrows=1, figsize=[7, 4])
    ax2 = ax.twinx()
    line1, = ax.plot(x_keVs, y_max, label='max spectrum', color='green')
    line2, = ax2.plot(x_keVs, y_sum, label='sum spectrum', color='blue')


    # annotate iron in max spectrum 
    ax.axvline(iron_xy[0], linestyle=':', color='r')
    ax.annotate(f'Iron Ka \n({iron_xy[0]:.3f} keV)', iron_xy, xytext=[20, -20], 
                textcoords='offset points', ha='left', 
                arrowprops=dict(arrowstyle="->", facecolor='grey'))
    ax.scatter(*iron_xy, s=40, facecolor=[1, 0.5, 0.5], edgecolor='green', zorder=10) 

    ax.set_title('Energy Calibration')
    ax.set_xlabel('Energy [keV]')
    ax.set_ylabel('Max Spectrum Intensity [counts]')


    # annotate compton peak in sum spectrum 
    ax2.annotate('Compton Peak\n(inelastic scattering)\n' + 
                r'$\theta$' + f'={int(theta)}' + r'$^{\circ}$', 
                compton_xy, xytext=[20, 20], 
                textcoords='offset points', ha='left', 
                arrowprops=dict(arrowstyle="->", facecolor='grey'))
    ax2.scatter(*compton_xy, s=40, facecolor='white', edgecolor='blue', zorder=10)       

    # annotate anode peak in sum spectrum 
    ax2.axvline(anode_xy[0], linestyle=':', color='r')
    ax2.annotate(f'Anode\nRhodium Ka \n({anode_xy[0]:.2f} keV)', anode_xy, xytext=[20, 20], 
                textcoords='offset points', ha='left', 
                arrowprops=dict(arrowstyle="->", facecolor='grey'))
    ax2.scatter(*anode_xy, s=40, facecolor=[1, 0.5, 0.5], edgecolor=[0, 0, 1], zorder=10)        

    ax2.set_ylabel('Sum Spectrum Intensity [counts]')

    ax.legend(handles=[line1, line2], loc='upper right')
    
    # force updating plot 
    plt.pause(0.2)
    fig.canvas.draw()
        
    # write energies 
    write = input('Write instrument energy calibration to datastack file [y/n]? ')

    if write == 'y': 
        print(f'\nWriting channel energies (keV) to: {datastack_file}')
        maxrf4u.append(x_keVs, ds.MAXRF_ENERGIES, datastack_file) 
        
        print(f'Also writing instrument Compton anode peak energy (keV) to: {datastack_file}')
        compton_peak_energy = x_keVs[compton_peak_i]
        maxrf4u.append(np.array([compton_peak_energy]), 'compton_peak_energy', datastack_file)
            
    return x_keVs


def detector_angle(keV0, keV1): 
    '''Calculate detector scatter angle `theta` (degrees) 
    
    From Compton peak energy `keV0` and anode energy `keV1`, assuming a single scattering event. 
    
    Returns: theta (degrees)'''
    
    assert keV0 < keV1, 'Compton energy should be less then anode energy'

    # constants 
    e = 1.602e-19 # electron charge
    pi = math.pi
    h = 6.62607e-34 # Planck constant 
    m0 = 9.10939e-31 # electron mass 
    c = 2.998e8 # speed of light 
    
    
    # convert keV to Joule 
    E0 = 1000 * keV0 * e 
    E1 = 1000 * keV1 * e 

    theta = (180 / pi) * np.arccos(1 - m0 * c**2 * (1/E0 - 1/E1))

    return theta 


def find_instrument_peaks(y_sum, prominence=0.1, tube_keV=40):  
    '''Locate key instrument peaks: 
    
              1) left hand sensor peak index  
              2) anode Compton peak index  
              3) right hand (rhodium) anode Ka peak index  
              
              in sum spectrum `y_sum`. 
    
    Assumes anode material is rhodium, and Compton peak energy is 
    first peak below (uncalibrated) 20 keV, based on tube keV.     
    
    Returns: [left_peak_i, compton_peak_i, right_peak_i] 
    '''
    
    # find all prominent peaks in sum spectrum 
    n_channels = len(y_sum) 
    x_indices = np.arange(n_channels)
    sum_peak_indices, shapes_dict = ssg.find_peaks(y_sum, prominence=prominence)
    
    # Sensor peak 
    # assume sensor peak index is first peak index in list    
    left_peak_i = sum_peak_indices[0] 
    
    # Compton peak (new approach)
    # assume that Compton peak is first peak below 20 keV 
    # in `tube_keV` based uncalibrated spectrum 
    # i.e. with energies ranging from 0 to `tube_keV` 
     
    x_approx_keVs = np.linspace(0, tube_keV, n_channels)
    peak_approx_keVs = x_approx_keVs[sum_peak_indices] # approximate peak keVs from tube keV 
    is_below20keV = peak_approx_keVs < 20 
    
    i = np.argmax(peak_approx_keVs[is_below20keV]) # peak number 
    compton_peak_i = sum_peak_indices[i]
    
    # Anode peak 
    # find anode peak channel index right next to Compton peak    
    right_peak_i = sum_peak_indices[list(sum_peak_indices).index(compton_peak_i) + 1] 
    
    return [left_peak_i, compton_peak_i, right_peak_i]
      



def compton_shift(keV_in, theta): 
    '''Compute Compton shift for photon energies `keV_in` and scatter angle `theta`. 
    
    Assuming single scattering. 
    
    Returns: keV_out'''
    

    # See: https://en.wikipedia.org/wiki/Compton_scattering 

    # constants 
    e = 1.602e-19 # electron charge
    pi = math.pi
    h = 6.62607e-34 # Planck constant 
    m0 = 9.10939e-31 # electron mass 
    c = 2.998e8 # speed of light

    # convert keV to Joule 
    E_in = 1000 * keV_in * e 
    
    # convert theta to radians 
    theta = pi * theta / 180
    
    E_out = E_in / (1 + (E_in / (m0 * c**2)) * (1 - np.cos(theta))) 
    
    keV_out = E_out / (1000 * e) 
    
    return keV_out 
    



# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"Layers": "10_storage.ipynb",
         "raw_to_datastack": "10_storage.ipynb",
         "tree": "10_storage.ipynb",
         "underscorify": "10_storage.ipynb",
         "append": "10_storage.ipynb",
         "repack": "10_storage.ipynb",
         "max_and_sum_spectra": "10_storage.ipynb",
         "DataStack": "10_storage.ipynb",
         "DATASTACK_EXT": "10_storage.ipynb",
         "L": "10_storage.ipynb",
         "calibrate": "20_calibration.ipynb",
         "detector_angle": "20_calibration.ipynb",
         "find_instrument_peaks": "20_calibration.ipynb",
         "compton_shift": "20_calibration.ipynb",
         "RHODIUM_Ka": "20_calibration.ipynb",
         "IRON_Ka": "20_calibration.ipynb",
         "ImageRegistrationHelper": "30_spatial-verus-spectral.ipynb",
         "warp": "30_spatial-verus-spectral.ipynb",
         "xy_to_latlon": "30_spatial-verus-spectral.ipynb",
         "latlon_to_xy": "30_spatial-verus-spectral.ipynb",
         "HotmaxAtlas": "40_hotmax.ipynb",
         "compute_hotmax_spectra": "40_hotmax.ipynb",
         "compute_hotmax_noise": "40_hotmax.ipynb",
         "get_patterns": "50_peak-pattern-puzzle.ipynb",
         "colorize": "50_peak-pattern-puzzle.ipynb",
         "plot_ptrn": "50_peak-pattern-puzzle.ipynb",
         "plot_patterns": "50_peak-pattern-puzzle.ipynb",
         "plot_puzzle": "50_peak-pattern-puzzle.ipynb",
         "get_instrument_pattern": "50_peak-pattern-puzzle.ipynb",
         "add_hotlines": "50_peak-pattern-puzzle.ipynb",
         "all_elements": "50_peak-pattern-puzzle.ipynb",
         "eoi": "50_peak-pattern-puzzle.ipynb",
         "get_continuum": "60_peakmaps.ipynb",
         "get_gaussians": "60_peakmaps.ipynb",
         "fit_spectrum": "60_peakmaps.ipynb",
         "get_spectral_slices": "60_peakmaps.ipynb",
         "compute_nmf_peakmaps": "60_peakmaps.ipynb",
         "multi_plot": "60_peakmaps.ipynb",
         "compute_nmf_element_maps": "60_peakmaps.ipynb",
         "gaussian_convolve": "90_xphysics.ipynb",
         "ElementLines": "90_xphysics.ipynb",
         "get_element_spectrum": "90_xphysics.ipynb",
         "get_element_spectra": "90_xphysics.ipynb",
         "UploadDir": "91_interactive-plotting.ipynb",
         "make_filenames": "91_interactive-plotting.ipynb",
         "make_gridbox_widget": "91_interactive-plotting.ipynb"}

modules = ["storage.py",
           "calibration.py",
           "warp.py",
           "hotmax.py",
           "peakpuzzle.py",
           "peakmaps.py",
           "xphysics.py",
           "interactive_visualization.py"]

doc_url = "https://fligt.github.io/maxrf4u/"

git_url = "https://github.com/fligt/maxrf4u/tree/main/"

def custom_doc_links(name): return None



__version__ = "0.1.11"

from .storage import *
from .calibration import * 
from .warp import * 
from .hotmax import * 
from .peakpuzzle import *
from .peakmaps import * 
from .xphysics import * 

from .interactive_visualization import * 

#from .segmentation import * 
from .speckles import * 

#from .fingerprints import * 






"""Lasagna versus spaghetti"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/30_spatial-verus-spectral.ipynb.

# %% auto 0
__all__ = ['ImageRegistrationHelper', 'warp', 'xy_to_latlon', 'latlon_to_xy']

# %% ../notebooks/30_spatial-verus-spectral.ipynb 35
from ipywidgets import Layout, HBox, VBox 
#import ipywidgets 

from ipyleaflet import Map, ImageOverlay, Marker, DivIcon, FullScreenControl, projections

#from ipyleaflet import Map, Marker, projections, ImageOverlay, FullScreenControl, DivIcon

from ipywidgets import Layout, GridBox, VBox, HBox, jsdlink 
from ipywidgets.embed import embed_minimal_html 
  
import io 
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np
import base64 
import skimage.transform as skt 
import imageio.v2 as imageio 
import skimage.transform as skt

import numpy as np
import cv2 

# %% ../notebooks/30_spatial-verus-spectral.ipynb 36
class ImageRegistrationHelper(object): 
    '''Create interactive image registration. '''
    
    def __init__(self, src_im, dst_im, src_points=None, dst_points=None, scale_down=True): 
        
         
        # image shapes 
        src_h, src_w = self.src_h, self.src_w = src_im.shape[0:2]
        dst_h, dst_w = self.dst_h, self.dst_w = dst_im.shape[0:2]
        
        
        # scale base64 image down to smallest width 
        if  scale_down is True: 
            max_width = min([src_w, dst_w]) 
        else: 
            max_width = None 
        
        # create base64 urls 
        src_url = _img_to_base64_url(src_im, max_width=max_width)
        dst_url = _img_to_base64_url(dst_im, max_width=max_width)
        
        # create custom colored icons
        w, h = 60, 20
        colors = ['orange', 'green', 'red', 'blue'] 
        markers_html = [f"""<svg width="60px", height="45px">
                <polygon points="20,25 40,25 30,40" style="fill:{col};stroke:white;stroke-width:1" />
                <text x="10" y="12" font-size="10" fill="black">Drag me</text>
                </svg>""" for col in colors] 
        icons = [DivIcon(html=html, bg_pos=(0, 0), icon_anchor=[30, 40], icon_size=[w, h]) for html in markers_html]

        # prepare src and dst markers 
        self._src_markers = [] 
        self._dst_markers = [] 
        
        src_locations = xy_to_latlon(imshape=[src_h, src_w], xy_points=src_points) 
        dst_locations = xy_to_latlon(imshape=[dst_h, dst_w], xy_points=dst_points) 

        # calculate map src and dst zoom levels to fit bounds (leaflet tiles are 256 pixels)
        # see: https://leafletjs.com/examples/zoom-levels/ 
        src_zoom = -max([np.log2(np.ceil(src_w / 256)), np.log2(np.ceil(src_h / 256))]) - 2
        dst_zoom = -max([np.log2(np.ceil(dst_w / 256)), np.log2(np.ceil(dst_h / 256))]) - 2 
        
        # create source maps for each corner 
        map_layout = Layout(height='200px', width='200px')
        src_maps = []
        for i, location in enumerate(src_locations):  
            m = Map(center=[src_h/2, src_w/2], zoom=src_zoom+1, crs=projections['Simple'],  layout=map_layout, 
                    scroll_wheel_zoom=True, min_zoom=src_zoom-2, interpolation='nearest')

            imo = ImageOverlay(url=src_url, bounds=[[-0.5, -0.5], [src_h - 0.5, src_w - 0.5]]) # bounds= SW NE corners
            mrk = Marker(location=location, draggable=True, title='Drag me to landmark', icon=icons[i])
            fsc = FullScreenControl()

            m.add(imo)
            m.add(fsc)
            m.add(mrk)

            self._src_markers.append(mrk)

            m.remove(m.layers[0]) # hack to remove world map 
            src_maps.append(m)
            
        # create dst maps for each corner 
        dst_maps = []    
        for j, location in enumerate(dst_locations):  
            m = Map(center=[dst_h/2, dst_w/2], zoom=dst_zoom+1, crs=projections['Simple'],  layout=map_layout, 
                    scroll_wheel_zoom=True, min_zoom=dst_zoom-2, interpolation='nearest')

            imo = ImageOverlay(url=dst_url, bounds=[[-0.5, -0.5], [dst_h - 0.5, dst_w - 0.5]]) # bounds= SW NE corners
            mrk = Marker(location=location, draggable=True, title='Drag me to landmark', icon=icons[j])
            fsc = FullScreenControl()

            m.add(imo)
            m.add(fsc)
            m.add(mrk)

            self._dst_markers.append(mrk)

            m.remove(m.layers[0]) # hack to remove world map 
            dst_maps.append(m)
        
        # sync map widgets centers to marker locations  
        # not exactly what I want, but close 
        # need probably something like a callback function on_drag or zoom  
        # too complicated for now 
        #if center_markers: 
        #    for  mrk, mp in zip(self._src_markers, src_maps): 
        #        jsdlink([mrk, 'location'],[mp, 'center'])
        #
        #    for  mrk, mp in zip(self._dst_markers, dst_maps): 
        #        jsdlink([mrk, 'location'],[mp, 'center'])
        #           
        # combine maps     

        src_hbox = HBox(src_maps)
        dst_hbox = HBox(dst_maps)
        
        self._vbox = VBox([src_hbox, dst_hbox])
        
    def show(self): 

        return self._vbox 
    
    
    def get_marker_coordinates(self, verbose=True): 
        '''Extract marker point coordinates from widget. '''
        
        src_locations = [mrk.location for mrk in self._src_markers]
        dst_locations = [mrk.location for mrk in self._dst_markers]
        
        src_points = latlon_to_xy([self.src_h, self.src_w], src_locations)
        dst_points = latlon_to_xy([self.dst_h, self.dst_w], dst_locations)
        
        src_points = np.round(src_points, decimals=2).tolist()
        dst_points = np.round(dst_points, decimals=2).tolist()
        
        if verbose: 
            print(f'src_points = {src_points}')
            print(f'dst_points = {dst_points}')
               
        return src_points, dst_points 
    
        
        
def _img_to_base64_url(img_data, max_width=None): 
    '''Filename or numpy array *img_data* is transformed into 256 bit color compressed base64 encoded url string. 
    
    To reduce image size specify thumbnail `max_width`. 
    
    Returns: url_string'''
        
    #try: 
    #    # better than matplotlib use skimage.io.imread to avoid float64 explosion...
    #    img = imageio.imread(img_data) # if img_data is an image file path 
    #except: 
    #    img = img_data # otherwise assume img_data is an image like numpy array 
    img = img_data 

    # rescaling image if width > max_width 
    # TODO: better testing 
    h, w = img.shape[0:2] 
    if max_width is not None:  
        if w > max_width:
            scale = max_width / w 
            # scalar image 
            if img.ndim == 2: 
                channel_axis = None
            else: 
                channel_axis = 2
            img = skt.rescale(img, scale, channel_axis=channel_axis) #, multichannel=True) deprecated 
    # apply color map if single channel image
    sm = cm.ScalarMappable()
    sm.set_clim(vmin=None, vmax=None)
    rgba = sm.to_rgba(img, bytes=True)

    # convert to base64 string via memory buffer 
    buff = io.BytesIO();
    imageio.imsave(buff, rgba, format='png')
    base64_string = base64.b64encode(buff.getvalue()).decode("ascii")
    url_string = f'data:image/png;base64,{base64_string}' 
    
    # let's close buffer just in case
    buff.close() 

    return url_string 

        
def warp(im_src, im_dst, pts_src, pts_dst, keep_scale=True, rgba=True, alpha_color=[1, 0, 0]): 
    '''Opencv based homographic registration. Can return transparent overlay (rgba). 
    
    Returns: *im_warped*, *extent* '''

    src_h, src_w = im_src.shape[0:2]
    dst_h, dst_w = im_dst.shape[0:2]
    
    if keep_scale: 
        scale = src_h / dst_h 
    else: 
        scale = 1

    # calculate homography 
    hom, status = cv2.findHomography(pts_src, scale * pts_dst)

    # Size is nearest integer scaled (width,height) of im_dst
    size = round(scale * dst_w), round(scale * dst_h)
    
    # warp im_src onto destination 
    im_warped = cv2.warpPerspective(im_src, hom, size) 
    
    # warp white onto destination to create mask 
    mask_src = np.ones([src_h, src_w])
    mask_warped = cv2.warpPerspective(mask_src, hom, size) 
    
    # calculate extent for plotting 
    h, w = im_dst.shape[0:2] 
    extent =  (0, w, h, 0) 
    
    # convert into rgba image  
    if rgba is True: 
        im_rgba = np.ones([h, w, 4])

        # for single channel image 
        if len(im_src.shape) == 2:
            # colorize
            im_rgba[:,:,0:3] = alpha_color[0:3] 
            #  make alpha layer 
            im_rgba[:,:,3] = im_warped 
            
        # otherwise assume image is rgb
        elif len(im_warped.shape) == 3: 
            
            # rgb channels
            im_rgba[:,:,0:3] = im_warped[:,:,0:3]
            # alpha channels 
            im_rgba[:,:,3] = mask_warped
       
        im_warped = im_rgba 
    
    return im_warped, extent  


# coordinate transformations 

def xy_to_latlon(imshape, xy_points=None): 
    '''Flip and exchange list with xy `points` into list of latitude-longitude locations. 
    
    If `xy_points` are not specified, locations of the four image corner are returned. 
    
    Returns: locations '''
    
    h, w = imshape 
    
    # use image corners (pixel centric coordinates) 
    if xy_points is None:     
        xy_points = [[0, 0], [w - 1, 0], [0, h - 1], [w - 1, h - 1]]
    
    xy_points = np.array(xy_points) # cast to numpy array if list
    xy_points[:,1] = h - 1 - xy_points[:,1] # flip y 
    
    locations = xy_points[:,::-1] # exchange x <-> y 
    
    locations = locations.tolist() # Marker() expects list instead of array 

    return locations 


def latlon_to_xy(imshape, locations): 
    '''Flip and exchange list with latitude-longitude `locations` into list of xy_points. 
    
    Returns: xy_points '''
    
    h, w = imshape 
    
    locations = np.array(locations) # cast to numpy array if list 
    locations[:,0] = h - 1 - locations[:,0] # flip y  
    
    xy_points = locations[:,::-1] # exchange x <-> y 

    return xy_points



"""The amazing possibilities of Jupyter ipywidgets and cloud storage"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/80_interactive-plotting.ipynb.

# %% auto 0
__all__ = ['UploadDir', 'make_filenames', 'make_gridbox_widget']

# %% ../notebooks/80_interactive-plotting.ipynb 25
import json 
import os 

import IPython 

from IPython.display import display
from IPython.display import Image
from IPython.display import HTML


from ipywidgets.embed import embed_minimal_html, embed_data 
from ipywidgets import Layout, HBox, VBox, GridBox, jslink, HTML 
from ipywidgets import Widget 

from ipyleaflet import (Map, projections, ImageOverlay, Rectangle, ZoomControl, FullScreenControl, 
                        DrawControl, WKTLayer, Popup)
import os 

import PIL
import urllib 
import numpy as np 
import matplotlib.pyplot as plt

# %% ../notebooks/80_interactive-plotting.ipynb 26
class UploadDir(): 
    
    def __init__(self, mount_dir, objnr, bucket_url, subdir='images'):
        '''Creates a standard upload (image) directory object. '''
        
        self._mount_dir = mount_dir
        self._objnr = objnr 
        self._subdir = subdir 
        self._bucket_url = bucket_url 
        
        # check mount 
        assert os.path.isdir(mount_dir), f'Directory {mount_dir} was not found. Forgot to mount?' 

        # create subdirectory if needed  
        self._img_dir = os.path.join(mount_dir, objnr, subdir) 
        os.makedirs(self._img_dir, exist_ok=True)
        
        self._images_url = os.path.join(bucket_url, objnr, subdir)
        
        print(f'Created online images folder: {self._images_url}')

        
    def imsave(self, img_list, filename_list, verbose=True): 
        '''Save image arrays `img_list` as `filename_list` in cloud storage.  

        Returns: img_url_list 
        ''' 
        
        print(f'Saving {len(img_list)} images to folder: {self._images_url}')
        
        img_url_list = []
        for img, filename in zip(img_list, filename_list):   
            filepath = os.path.join(self._img_dir, filename) 
            
            # save to rclone mounted image folder 
            plt.imsave(filepath, img) 
            
            img_url = os.path.join(self._bucket_url, self._objnr, self._subdir, filename)
            img_url_list.append(img_url)
            
        if verbose: 
            print('\n')
            for img_url in img_url_list: 
                print(img_url)

        return img_url_list 
    
    
    def listdir(self, filepath=True): 
        '''List local (mounted) file paths for images subfolder. '''
        
        filenames = os.listdir(self._img_dir)
        
        # prefix filepath 
        if filepath:  
            filenames = [os.path.join(self._img_dir, fn) for fn in filenames] 
            
        return filenames 
    
    def listurls(self): 
        '''Construct image urls'''
        
        filename_list = os.listdir(self._img_dir)
        
        img_url_list = []
        for filenames in filename_list: 
            img_url = os.path.join(self._bucket_url, self._objnr, self._subdir, filename)
            img_url_list.append(img_url)
            
        return img_url_list 
    
    def export_interactive_html(self, widget, viztype): 
        '''Save `widget` as interactive html page in cloud storage.
        
        Returns: html_url'''
        
        # create html subdirectory if needed  
        html_dir = os.path.join(self._mount_dir, self._objnr, 'html') 
        os.makedirs(html_dir, exist_ok=True)
        
        html_filename = f'{self._objnr}_{viztype}.html' 
        html_filepath = os.path.join(html_dir, html_filename)
        html_url = os.path.join(self._bucket_url, self._objnr, 'html', html_filename)

        print(f'Saving interactive html to cloud storage...')
        embed_minimal_html(html_filepath, widget)
        print('Click link to load the interactive visualization (opens a separate page).\n')

        print(html_url)
        
        return html_url 
        
    
    
def make_filenames(objnr, viztype, titles, ext): 
    '''Creates standard filenames. 
    
    Returns: filenames
    '''
    
    filenames = []
    for t in titles: 
        fname = f'{objnr}_{viztype}_{t}.{ext}'
        filenames.append(fname)
    
    return filenames




def make_gridbox_widget(img_urls, titles, shape=None): 
    '''Creates multi-image interactive synchronized viewer. '''
    
    # close all widgets to avoid ever growing html export file size  
    Widget.close_all()
    
    # if shape is not specified 
    # assume all images map onto shape of first image 
    if shape is None: 
        shape = np.array(PIL.Image.open(urllib.request.urlopen(img_urls[0]))).shape[0:2] # height, width 
        h, w = shape  

    map_layout = Layout(width='20vw', height='20vw')
    vbox_layout = Layout(width='20.5vw', height='20vw')
    grid_layout = Layout(grid_template_columns="repeat(5, 20.2vw)") 
    
    # make map widgets  
    
    map_widgets = [] 
    for url in img_urls:

        m = Map(center=[h/2, w/2], zoom=-4, crs=projections['Simple'], layout=map_layout, 
                    scroll_wheel_zoom=True, min_zoom=-5)#, interpolation='nearest')
        
        imo = ImageOverlay(url=url, bounds=[[0, 0], [h, w]]) # bounds= SW NE corners
        fsc = FullScreenControl()
        
        m.add(imo)
        m.add(fsc)
        m.remove(m.layers[0]) # hack to remove world map 
        
        map_widgets.append(m)
        
    # sync map widgets centers and zoom levels via browser javascript 
    first, rest = map_widgets[0], map_widgets[1:]
    for r in rest: 
        jslink([first, 'center'],[r, 'center'])
        jslink([first, 'zoom'],[r, 'zoom'])
    
    # titles 
    
    title_widgets = []
    for t in titles: 
        
        tw = HTML(f'<center><bf>{t}</bf></center>') # ? 
        title_widgets.append(tw)
    
    # combine titles and maps 
    vboxes = []
    for title_widget, map_widget in zip(title_widgets, map_widgets): 
        
        vbox = VBox([title_widget, map_widget], layout=vbox_layout)
        vboxes.append(vbox)
    
    gridbox = GridBox(vboxes, layout=grid_layout)
    
    return gridbox 

                       



"""Dealing with mixed signals"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/60_peakmaps.ipynb.

# %% auto 0
__all__ = ['get_continuum', 'get_gaussians', 'fit_spectrum_peaks', 'compute_nmf_peakmaps', 'multi_plot',
           'compute_nmf_element_maps']

# %% ../notebooks/60_peakmaps.ipynb 30
from dask import delayed 
from dask.diagnostics import ProgressBar 
import maxrf4u 
import numpy as np 
import scipy.interpolate as sip 
import scipy.signal as ssg 
import matplotlib.pyplot as plt 
from IPython.display import SVG 

import sklearn.decomposition as skd 

# %% ../notebooks/60_peakmaps.ipynb 31
def get_continuum(datastack_file): 
    '''Compute continuum baseline from sum spectrum.
    
    Uses rolling ball filter to remove peaks from sum spectrum. 
    
    Returns: y_continuum
    '''
    
    ds = maxrf4u.DataStack(datastack_file) 
    x_keVs = ds.read('maxrf_energies')
    y_sumspectrum = ds.read('maxrf_sumspectrum') 
    
    # rolling ball filter
    y_continuum = maxrf4u.hotmax._baseline(x_keVs, y_sumspectrum) # radius=200 
    # repeat with smaller ball to cut sharp corners of Compton 
    y_continuum =  maxrf4u.hotmax._baseline(x_keVs, y_continuum, radius=40)
    
    return y_continuum 


def _gaussian(x, x0, sigma): 
    '''Normal distribution around `x0` with standard deviation `sigma`.'''
    
    y = np.exp(-0.5 * ((x - x0) / sigma)**2)
    
    return y 


def _fit_gaussian(x, y, peak_idx, rel_height=0.2, baseline=None): 
    '''Fit single gaussian distribution at `rel_height`. 
    
    Returns: `y_gauss`, `baseline`
    '''
    
    # just a single peak 
    [widths], [width_height], [left_ips], [right_ips] = ssg.peak_widths(y, [peak_idx], rel_height=rel_height)
    
    # create energy interpolation 
    keV_ipol = sip.interp1d(np.arange(len(x)), x)
     
    left_x = keV_ipol(left_ips)
    right_x = keV_ipol(right_ips)
    
    # pick smallest half width at rel_height 
    dx_left = left_x - x[peak_idx] 
    dx_right = right_x - x[peak_idx] 
    
    dx = sorted([dx_left, dx_right])[0] 
    
    # create baseline if None   
    if baseline is None:  
        baseline = np.zeros_like(y) 
        
    y_norm = (width_height - baseline[peak_idx]) / (y[peak_idx] - baseline[peak_idx]) 
    
    # assert that y_norm is positive 
    assert y_norm > 0, f"Can't fit Gaussian for peak below baseline for peak index {peak_idx}."   
        
    # calculate corresponding sigma 
    sigma = np.sqrt(-dx**2 / (2 * np.log(y_norm)))
    
    # calculate gaussian with baseline 
    y_gauss = (y[peak_idx] - baseline[peak_idx]) * _gaussian(x, x[peak_idx], sigma) + baseline
    
    return y_gauss, baseline 


def get_gaussians(datastack_file, tail_clip=0.05, norm=True): 
    '''Fit Gaussian peak profiles to hotmax peaks in hotmax atlas. 

    Returns: y_gauss_list
    '''

    # read stuff from datastack 
    ds = maxrf4u.DataStack(datastack_file) 

    x_keVs = ds.read('maxrf_energies') 
    hotmax_spectra = ds.read('hotmax_spectra')
    hotmax_baselines = ds.read('hotmax_baselines') 
    peak_idxs_list = ds.read_list('hotmax_peak_idxs_list')

    # get slices by fitting gaussian to corresponding hotmax spectrum and baseline 
    y_gauss_list = []

    # Create tail clipped gaussians for ragged peak indexes list 
    for n, peak_idxs in enumerate(peak_idxs_list):  

        y_hot = hotmax_spectra[n]
        baseline = hotmax_baselines[n] 

        # iterate over (possibly multiple) peaks in each hotmax spectrum 
        for idx in peak_idxs: 

            y_gauss_fit, baseline = _fit_gaussian(x_keVs, y_hot, idx, baseline=baseline) 
            y_gauss_flat =  y_gauss_fit - baseline 

            if norm is True: 
                y_gauss_flat = y_gauss_flat / y_gauss_flat.max()

            y_gauss_list.append(y_gauss_flat)
            
    return y_gauss_list 


def fit_spectrum_peaks(y, datastack_file): 
    '''Fit Gaussian peak shapes to single spectrum `y`.
    
    Uses NMF with fixed Gaussian component vectors. 
    
    Returns: weighed_components_list 
    '''

    X = y.reshape([1, -1]) # make single spectro into two dimenional array 
    

    y_continuum = get_continuum(datastack_file)
    y_gauss_list = get_gaussians(datastack_file, norm=False)

    # create component vectors 
    H = np.array([y_gauss for y_gauss in y_gauss_list]).astype(np.float32)
    # add continuum as component
    H = np.r_[H, y_continuum[None,:].astype(np.float32)]
    
    n_components, n_channels = H.shape

    W, H, n_iter = skd.non_negative_factorization(X, H=H, n_components=n_components, update_H=False)
    
    # not neede here, but just a reminder  
    # X_nmf = W @ H 
    # y_nmf = X_nmf.flatten() 

    # transpose weights for easy multiplication  
    w =  W.T 
    
    return w, H 
        

    
def _add_hotlines_ticklabels(datastack_file, ax, vlines=True, clip_vline=True): 
    '''Utility function. Adds hotlines and tick labels to plot `ax`. 
    
    '''
    # read hotlines data 
    ds = maxrf4u.DataStack(datastack_file) 
    x_keVs = ds.read('maxrf_energies')
    peak_idxs = ds.read('hotmax_peak_idxs_flat') 
    
    secax = ax.secondary_xaxis('top')

    secax.set_xticks(x_keVs[peak_idxs]) 
    
    ymin, ymax = ax.get_ylim() 
    
    # include vlines in plot 
    if vlines: 
        # clip vlines at y=0
        if clip_vline: 
            ax.vlines(x_keVs[peak_idxs], 0, 1.2*ymax, color='r', alpha=0.2, zorder=9-30)
            ax.set_ylim(ymin, ymax)

        # do not clip vlines 
        else: 
            for x in x_keVs[peak_idxs]:
                ax.axvline(x, color='r', alpha=0.2, zorder=9-30)

    secax.set_xticks(x_keVs[peak_idxs])
    secax.set_xticklabels(range(len(peak_idxs)), fontsize=6, color='r') 
    secax.tick_params(color=[1, 0.5, 0.5], pad=0) 
    
    return ax 


def _get_spectral_slices(cube): 
    '''Low level bookkeeping to compute slice indices for spectral chunks from dask array `cube`. 
       
    Returns: row_slices, column_slices  
    '''
    
    h, w, d = cube.shape 

    row_chunks, column_chunks, spectral_chunks = cube.chunks

    row_idxs = [0] 
    for c in row_chunks: 
        row_idxs.append(row_idxs[-1] + c)

    row_slices = [[row_idxs[i-1], row_idxs[i]] for i in range(1, len(row_idxs))]

    # columns 
    column_idxs = [0] 
    for c in column_chunks: 
        column_idxs.append(column_idxs[-1] + c)

    column_slices = [[column_idxs[i-1], column_idxs[i]] for i in range(1, len(column_idxs))] 

    return row_slices, column_slices 



def compute_nmf_peakmaps(datastack_file): 
    '''Use nonnegative matrix factorization to compute peak maps spectral data in `datastack_file`. 
    
    Asks to write result to datastack file. 
    
    Returns: peak_maps 
    '''
    
    ds = maxrf4u.DataStack(datastack_file)
    cube = ds.read('maxrf_cube', compute=False)
    h, w, d = cube.shape 
    
    
    # prepare gaussian component vectors 
    y_continuum = get_continuum(datastack_file)
    y_gauss_list = get_gaussians(datastack_file)

    H = np.array([y_gauss for y_gauss in y_gauss_list]).astype(np.float32)
    # add continuum as component
    H = np.r_[H, y_continuum[None,:].astype(np.float32)]

    n_components, n_channels = H.shape

    #initialize peak map stack  
    peak_map_stack = np.zeros([h, w, n_components], dtype=np.float32)

    # prepare chunk bookkeeping 
    row_slices, column_slices = _get_spectral_slices(cube)

    n_chunks = len(row_slices) * len(column_slices)

    n = 1 
    print('Please wait a few minutes while preparing your peak maps. ')
    for ri, rj in row_slices: 

        for ci, cj in column_slices: 
            print(f'Computing factorization for spectral chunk {n}/{n_chunks}...', end='\r')

            spectral_chunk = cube[ri:rj, ci:cj, :].compute()
            h, w, d = spectral_chunk.shape
            X = spectral_chunk.reshape([h*w, d])

            W, H, n_iter = skd.non_negative_factorization(X, H=H, n_components=n_components, update_H=False) 

            W_img = W.reshape([h, w, -1]) 

            peak_map_stack[ri:rj, ci:cj] = W_img 
            
            n = n + 1

    peak_maps = peak_map_stack.transpose([2, 0, 1])
    
    # user input           
    write = input('Write peak maps factorization to datastack file [y/n]? ')
         
    if write == 'y': 
        maxrf4u.append(peak_maps, 'nmf_peakmaps', datastack_file)
        maxrf4u.append(H, 'nmf_gausscomponents', datastack_file) 
        
        print(f'\nSaved NMF peak maps data to: {datastack_file}')
    
    return peak_maps 


def multi_plot(*images, hot_pixel=None, titles=None, roi_list=None, axis_off=False,
               sharex=True, sharey=True, vmin=None, vmax=None, cmap='viridis',
               fontsize='medium', zoom_xyc=None, zoom_half_wh=[100, 100], svg=False):
    '''Inspect multiple images simultaneously...
    
    Fold along multiple rows if n > 4'''
    
    if svg:
        plt.ioff()
    
    nrows_max = 4
    n_img = len(images)
    nrows = (n_img // nrows_max) # completely filled rows
    rest = n_img % nrows_max
    if rest != 0:
        nrows = nrows + 1
    if n_img <= nrows_max:
        ncols = n_img
    else:
        ncols = nrows_max
    figsize = [9, 5 + 1.3 * (nrows -1)]
    fig, axs = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize, squeeze=False, sharex=sharex, sharey=sharey)
    for i, img in enumerate(images):
        axs.flatten()[i].imshow(img, cmap=cmap, vmin=vmin, vmax=vmax)
        if hot_pixel is not None:
            hot_yi, hot_xi, hot_zi = hot_pixel
            axs.flatten()[i].scatter(hot_xi, hot_yi, color='r')
        if zoom_xyc is not None:
            xc, yc = zoom_xyc
            w_ha, h_ha = zoom_half_wh
            axs.flatten()[i].set_xlim(xc - w_ha, xc + w_ha)
            axs.flatten()[i].set_ylim(yc + h_ha, yc - h_ha)
        if roi_list is not None:
            add_roi_patches(axs.flatten()[i], roi_list)
    if titles is not None:
        for i, t in enumerate(titles):
            axs.flatten()[i].set_title(t, fontsize=fontsize)
    # seems to have no effect:
    for i in range(n_img, nrows * ncols):
        axs.flatten()[i].axis('off')
    # therefore trying this:
    if axis_off:
        axs_flat = axs.flatten()
        for ax in axs_flat:
            ax.set_axis_off()
    fig.subplots_adjust(hspace=0.1, wspace=0.03)
    plt.tight_layout()
    
    
    # larger image in documentation 
    if svg: 
        plt.savefig('plot.svg')
        fig.clear()
        plt.ion()
        return SVG('plot.svg')
    
    return fig, axs


def compute_nmf_element_maps(datastack_file, elements_unsorted, excitation_energy_keV=25): 
    '''Compute element maps for `elements` list. 
    
    Based on NMF factorization. Requires previously computed peak maps stored in `datastack_file`. 
    Sorts elements according to strongest (a.k.a. alpha) peak. Ask user to save results. 
    
    Returns: elements, element_maps
    '''

    ds = maxrf4u.DataStack(datastack_file)
    
    x_keVs = ds.read('maxrf_energies')

    nmf_gausscomponents_plus_continuum = ds.read('nmf_gausscomponents')
    nmf_gausscomponents = nmf_gausscomponents_plus_continuum[0:-1]

    # the data X that we want to factorize first is the element_spectra
    
    print('Preparing peaks to elements transformation matrix...')
    
    elements, element_spectra = maxrf4u.get_element_spectra(elements_unsorted, x_keVs, 
                                                            excitation_energy_keV)
    
    X = element_spectra.astype(np.float32) 
    H = nmf_gausscomponents 
    n_components, n_channels = H.shape 
    W_elem, H, n_iter = skd.non_negative_factorization(X, H=H, n_components=n_components, update_H=False) 

    # Ok, this might be slightly confusing, but now this weights matrix becomes the new components matrix. 

    peak_maps = ds.read('nmf_peakmaps')

    # Need to transpose first, then temporarily flatten.

    peak_maps_stack = peak_maps.transpose([1, 2, 0])

    h, w, d = peak_maps_stack.shape
    
    # this is actual the NMF transformation of peak maps into element maps  
    
    X = peak_maps_stack.reshape([h * w, d])  # flatten 
    X = X[:,0:-1] # get rid of continuum layer 

    H = W_elem 
    n_components, n_peaks = H.shape 
    
    print(f'Computing element maps for {elements}...')

    W2, H, n_iter = skd.non_negative_factorization(X, H=H, n_components=n_components, update_H=False)
    
    print('Ready computing!\n')

    # Ok, unflatten and transpose again. 

    element_maps = W2.reshape([h, w, -1]).transpose([2, 0, 1])
    
    # user input           
    write = input('Write NMF element maps and factorization data to datastack file [y/n]? ')
         
    if write == 'y': 
        maxrf4u.append(element_maps, 'nmf_elementmaps', datastack_file)
        maxrf4u.append(np.array(elements), 'nmf_elements', datastack_file) 
        maxrf4u.append(W_elem, 'nmf_peaks2elements_matrix', datastack_file)
        
        print(f'\nSaved NMF element maps and factorization data to: {datastack_file}')

    return elements, element_maps 



"""Cherry picking the best spectra!"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/40_hotmax.ipynb.

# %% auto 0
__all__ = ['HotmaxAtlas', 'compute_hotmax_spectra', 'compute_hotmax_noise', 'compute_subpeaks']

# %% ../notebooks/40_hotmax.ipynb 39
import maxrf4u 
import scipy.signal as ssg 
import numpy as np 
import time 
import matplotlib.pyplot as plt 
from dask.diagnostics import ProgressBar 
import scipy.ndimage.morphology as morph 
import scipy.interpolate as sip 
from IPython.display import SVG 
from itertools import chain 

# %% ../notebooks/40_hotmax.ipynb 40
class HotmaxAtlas(): 
    
    def __init__(self, datastack_file, prominence=0.2): 

        # add datastack name as attribute 
        self.datastack_file = datastack_file 
        
        # read from datastack file (arrays)
        ds = maxrf4u.DataStack(datastack_file)
        self.x_keVs = ds.read('maxrf_energies') 
        self.y_max = ds.read('maxrf_maxspectrum') 
        self.hotmax_spectra = ds.read('hotmax_spectra')
        self.baselines = ds.read('hotmax_baselines')
        self.noiselines = ds.read('hotmax_noiselines') 
        
        self.n_spectra = len(self.hotmax_spectra) 
        self.submax_spectrum = np.max(self.hotmax_spectra, axis=0) # core max spectrum
        
        # read from datastack file (ragged lists) 
        self.hotmax_peak_idxs_list = ds.read_list('hotmax_peak_idxs_list')
        self.hotmax_subpeak_idxs_list = ds.read_list('hotmax_subpeak_idxs_list')  

        # flatten ragged hotmax peak indexes list 
        self.hotmax_peak_idxs_flat = [] 
        for idxs in self.hotmax_peak_idxs_list: 
            self.hotmax_peak_idxs_flat.extend(idxs)
        
         
    def plot_spectrum(self, n, ax=None, legend=False, headspace=1, footspace=0.1, 
                      hotlines_ticklabels=True, tight_layout=False, xlim_padding=2): 
        
        if ax is None: 
            fig, ax = plt.subplots(figsize=[9, 3])
        
        # hotmax spectrum 
        ax.plot(self.x_keVs, self.hotmax_spectra[n], 
                zorder=-1, label='hotmax spectrum')
        
        # maxspectrum 
        ax.fill_between(self.x_keVs, self.y_max, color='r', alpha=0.15, zorder=8-30, label='max spectrum')
        ax.fill_between(self.x_keVs, self.submax_spectrum, color='r', alpha=0.1)
        
        # noise envelope 
        ax.fill_between(self.x_keVs, self.baselines[n], self.noiselines[n], 
                        color=[0.5, 0.9, 0.5], zorder=10-30, alpha=0.5, label='Poisson noise envelope')
        
        # fill subpeaks 
        ax.fill_between(self.x_keVs, self.noiselines[n], self.hotmax_spectra[n], 
                        where=self.hotmax_spectra[n] > self.noiselines[n],
                        color='b', alpha=0.5, zorder=8-30)

        # hotmax peaks (red square markers)
        hmp_x = self.x_keVs[self.hotmax_peak_idxs_list[n]]
        hmp_y = self.hotmax_spectra[n][self.hotmax_peak_idxs_list[n]]
        ax.scatter(hmp_x, hmp_y, marker='s', zorder=1, edgecolor='r', 
                   facecolor='w', label='hotmax peak(s)') 

        # (puzzle) sub peaks (round markers)
        subpeaks_x = self.x_keVs[self.hotmax_subpeak_idxs_list[n]]
        subpeaks_y = self.hotmax_spectra[n, self.hotmax_subpeak_idxs_list[n]]
        ax.scatter(subpeaks_x, subpeaks_y, edgecolor='b', alpha=1, zorder=0, facecolor='w', label='puzzle peaks')
        
        # annotate puzzle peaks 
        subpeak_labels = [f'({i})' for i, _ in enumerate(self.hotmax_subpeak_idxs_list[n])]
        ann_list = []
        for i, plabel in enumerate(subpeak_labels):  
            subpeak_xy = subpeaks_x[i], subpeaks_y[i] 
            ann = ax.annotate(plabel, subpeak_xy, xytext=[0, 10], color='b', 
                              textcoords='offset points', ha='center')
            ann_list.append(ann)
 

        # plot limits 
        ymax = 1.15 * self.hotmax_spectra[n].max() # add space for peak labels 
        ymin = -ymax / 5 
        ax.set_ylim(footspace * ymin, headspace * ymax)
        
        if xlim_padding != None: 
            xlim = self.x_keVs[max(self.hotmax_peak_idxs_list)[0]] + xlim_padding
            ax.set_xlim(-1, xlim)
        
        # remove negative yticks from footspace  
        yticks = ax.get_yticks()
        is_positive = yticks >= 0
        ax.set_yticks(yticks[is_positive])
        # again why ?? 
        ax.set_ylim(footspace * ymin, headspace * ymax)
        
        # plot text label 
        ax.text(0.995, 0.96, f'#{n}', ha='right', va='top', transform=ax.transAxes)
        
        # NEED TO ADJUST TO RAGGED LIST 
        # plot lines in pattern overview for all hotmax peaks  
        lines_x = self.x_keVs[self.hotmax_peak_idxs_flat]
        ax.vlines(lines_x, ymin=0, ymax=2*ymax*headspace, color='r', alpha=0.2, zorder=9-30)
        
        
        # add legend and labels etcetera if standalone 
        if legend:    
            ax.set_xlabel('Energy [keV]')
            ax.set_ylabel('Intensity [#counts]')
            ax.legend(loc='upper left', bbox_to_anchor=(1.01, 0.98)) 
        
        if tight_layout: 
            plt.tight_layout()
        
        return ax, ann_list 
    
        
    def plot_spectra(self, svg=True): 
        
        if svg: 
            plt.ioff()
        
        fig, axs = plt.subplots(nrows=self.n_spectra, figsize=[9, self.n_spectra], sharex=True)
        
        axs[0].set_title(f'Hotmax spectra')
  
        # set plot limits    
        xlim = self.x_keVs[max(self.hotmax_peak_idxs_list)[0]] + 2
        
        for n, ax in enumerate(axs): 
            self.plot_spectrum(n, ax=ax)
            
            # limits 
            ylim = 1.5 * self.hotmax_spectra[n].max()
            ax.set_ylim(-1, ylim)
            ax.set_xlim(-1, xlim)
            #ax.text(0.95, 0.95, f'{n}', ha='right', va='top', transform=ax.transAxes)
        
        ax.set_xlabel('Energy [keV]')
        
        plt.tight_layout() 
        plt.subplots_adjust(hspace=0.1)
        
        
        if svg: 
            plt.savefig('plot.svg') 
            fig.clear()
            plt.ion()
            return SVG('plot.svg')

        
        
def compute_hotmax_spectra(datastack_file, prominence=0.35): 
    '''Collect hotmax pixels and corresponding spectra from *datastack_file*. 
    
    Ask user confirmation for saving result to datastack file. 
    '''
   
    # read max spectrum 
    ds = maxrf4u.DataStack(datastack_file)
    x_keVs = ds.read(ds.MAXRF_ENERGIES)
    y_max = ds.read(ds.MAXRF_MAXSPECTRUM)

    # locate max spectrum peaks in max spectrum 
    peak_indices, _ = ssg.find_peaks(y_max, prominence=prominence)
    peaks_x = x_keVs[peak_indices]
    peaks_y = y_max[peak_indices]
    
    # plot max spectrum with peaks 
    
    #plt.close('all')
    #plt.ion() # important for updating plots with fig.canvas.draw() 
    
    fig, ax = plt.subplots(figsize=[9, 4])

    #fig.canvas.draw()   
    ax.fill_between(x_keVs, y_max, color='r', alpha=0.3)
    ax.scatter(peaks_x, peaks_y, facecolor='w', edgecolor='r')  
    ax.set_xlim(peaks_x[0] - 0.5, peaks_x[-1] + 5) 
    ax.set_ylim(-5, 1.15 * np.max(peaks_y))
    
    #plt.pause(0.3) 
    #fig.canvas.draw() 
    #fig.canvas.flush_events()
    
    for i, peak_xy in enumerate(zip(peaks_x, peaks_y)): 
        ax.annotate(f'[{i}]', peak_xy, xytext=[0, 10], color='r', 
                        textcoords='offset points', ha='center')
    ax.set_xlabel('Energy [keV]')
    ax.set_ylabel('Max spectrum intensity [counts]')
    ax.set_title(f'Found {len(peak_indices)} peaks with prominence > {prominence}')
    ax.grid()

    #plt.pause(0.3) 
    #fig.canvas.draw() 
    #fig.canvas.flush_events()
 
    
    # read corresponding channel maps for all max peak indices 
    print('Step 1/3: Reading hot max channel maps...')
    with ProgressBar(): 
        channel_maps = ds.maxrf_cube[:,:,peak_indices].compute() 

    channel_maps = channel_maps.transpose([2, 0, 1])

    # locate hot max pixels in channel maps 
    print('Step 2/3: Locating hot max pixels...')
    hot_pixels = []
    for k, c_map in enumerate(channel_maps): 
        hot_i, hot_j = np.argwhere(c_map == peaks_y[k]).flatten()
        hot_pixels.append([hot_i, hot_j, peak_indices[k]]) # xy

    hot_pixels = np.array(hot_pixels)  
    
    # need to deduplicate identical spectra (i.e. superimposed hotmax pixels)
    
    hotmax_spots, hotmax_peak_idxs_list = _group_superimposed(hot_pixels)
    

    # read spectrum for each hot max pixel  
    hotmax_spectra = [] 
    for n, [i, j] in enumerate(hotmax_spots):
        print(f'Step 3/3: Reading hot max spectrum {n}/{len(hotmax_spots) - 1}...', end='\r')
        spectrum = ds.maxrf_cube[i,j,:].compute()
        hotmax_spectra.append(spectrum)
        
        ax.plot(x_keVs, spectrum, color='b', linewidth=0.5, alpha=0.5) 
        ax.fill_between(x_keVs, spectrum, color='b', linewidth=0.5, alpha=0.3) 
        
        # DEPRECIATING MASK, INSTEAD USING RAGGED LIST  
        
        #x = x_keVs[hotmax_mask[n]]
        #y = spectrum[hotmax_mask[n]] 
        
        x = x_keVs[hotmax_peak_idxs_list[n]]
        y = spectrum[hotmax_peak_idxs_list[n]]
        ax.scatter(x, y, facecolor='b', edgecolor='r') 

        # force updating plot 
        #plt.pause(0.2) 
        #fig.canvas.draw()
        #fig.canvas.flush_events()
        
    print(f'Step 3/3: Ready with reading hot max spectra. ')

    hotmax_spectra = np.array(hotmax_spectra)   
    
    # user input           
    write = input('Write hotmax spectra, spots and peak indices to datastack file [y/n]? ')
         
    if write == 'y': 
        maxrf4u.append(hotmax_spectra, ds.HOTMAX_SPECTRA, ds.datastack_file)
        maxrf4u.append(hotmax_spots, 'hotmax_spots', ds.datastack_file) 
        maxrf4u.append_list(hotmax_peak_idxs_list, 'hotmax_peak_idxs_list', ds.datastack_file) 

        # also create a flat array of hotmax peak idxs  
        maxrf4u.append(hot_pixels[:, 2], 'hotmax_peak_idxs_flat', ds.datastack_file)
        

        print(f'\nSaved hotmax data to: {ds.datastack_file}')
        
    # force updating plot 
    #plt.pause(0.3) 
    #fig.canvas.draw()
    #fig.canvas.flush_events()
    
    return  



def compute_hotmax_noise(datastack_file, radius=200, alpha=0.6, beta=0.1): 
    '''Utility function to get baselines and noiselines from hotmax spectra in *datastack_file*.  
    
    Use once to compute base noise envelopes. Ask user confirmation for saving result to datastack file. '''
    
    ds = maxrf4u.DataStack(datastack_file)
    
    x_keVs = ds.read(ds.MAXRF_ENERGIES)
    hotmax_spectra = ds.read(ds.HOTMAX_SPECTRA)
        
    baseline_spectra = []  
    noiseline_spectra = []
    
    for i, spectrum in enumerate(hotmax_spectra): 
        
        print(f'Computing noise envelope for hotmax spectrum {i+1}/{len(hotmax_spectra)}...', end='\r')
        
        baseline, lowest_indices = _baseline(x_keVs, spectrum, radius=radius, return_indices=True)
        baseline_spectra.append(baseline)
        
        noiseline = _noiseline(baseline, alpha=alpha, beta=beta) 
        noiseline_spectra.append(noiseline) 
        
    print(f'Ready computing {len(hotmax_spectra)} noise envelopes.                           \n')
    
    # user input           
    write = input('Write hotmax baselines and noiselines to datastack file [y/n]? ')
         
    if write == 'y': 
        maxrf4u.append(baseline_spectra, ds.HOTMAX_BASELINES, ds.datastack_file)
        maxrf4u.append(noiseline_spectra, ds.HOTMAX_NOISELINES, ds.datastack_file) 

        print(f'\nSaved hotmax noise data to: {ds.datastack_file}')
    
    return 

def compute_subpeaks(datastack_file, prominence=0.2): 
    '''Find subpeaks in hotmax spectrum that exceed the noise envelope.
    
    Writes result to datas
    
    Returns: list of hotmax sub peak indices'''
    
    subpeak_idxs_list = []
    
    ds = maxrf4u.DataStack(datastack_file)
    hotmax_spectra = ds.read('hotmax_spectra')
    noiselines = ds.read('hotmax_noiselines')
    
    for y_hot, noiseline in zip(hotmax_spectra, noiselines): 
    
        # clip 
        above_noise = y_hot - noiseline 
        above_noise[above_noise < 0] = 0 

        peak_idxs, _ = ssg.find_peaks(above_noise, prominence=prominence) 

        # sort by peak height 
        peak_heights = above_noise[peak_idxs]
        peak_order = np.argsort(peak_heights)[::-1] 
        peak_idxs = peak_idxs[peak_order] 

        # hack to correct peak positions due to baseline tilting 
        peak_idxs = _untilt(peak_idxs, y_hot, delta=5) 
        
        # append 
        subpeak_idxs_list.append(peak_idxs)
         
    # user input           
    write = input('Write hotmax baselines and noiselines to datastack file [y/n]? ')    
    
    if write == 'y': 
         
        datapath = 'hotmax_subpeak_idxs_list'
        maxrf4u.append_list(subpeak_idxs_list, datapath, datastack_file)       
        print(f'\nSaved hotmax subpeak indexes list to: {datastack_file}')
        
    
    return subpeak_idxs_list 



def _baseline(x_keVs, spectrum, radius=200, return_indices=False): 
    '''Calculate rolling ball baseline for *spectrum*.  
    
    Uses Euclidian distance map to select only the lowest surges that touch the rolling ball.'''

    spectrum_norm = spectrum / spectrum.max()
    
    w = len(spectrum) 
    h = w // 2

    # initialize ramp with normalized spectrum in top half  
    ramp = np.ones([h, w]) * np.linspace(1, -1, h)[:, None]
    
    # create boolean mask from spectrum 
    spectrum_mask = ramp < spectrum_norm 
    
    # create distance map for spectrum mask. 
    dist_map, index_images = morph.distance_transform_edt(spectrum_mask, return_indices=True)

    # we only need the second (x-axis) channel index image
    _, index_img = index_images 

    # roll the ball 
    ball_mask = dist_map < radius 
    ball_profile = ball_mask.sum(axis=0) 

    # get touching surges indices 
    lowest_indices = list(set(index_img[ball_profile, range(w)]))
    
    # create linear interpolation between the selected lowest surges (valleys)
    # added fill_value=0 to avoid Nan's if extrapolating loose ends to avoid Nan's in baseline 
    interpolation = sip.interp1d(x_keVs[lowest_indices], spectrum[lowest_indices], bounds_error=False, fill_value=0)
    baseline = interpolation(x_keVs) 
    
    # valley indices 
    if return_indices: 
        return baseline, lowest_indices 
    
    return baseline 

def _noiseline(baseline, alpha=0.6, beta=0.1): 
    '''Estimate Poisson noise line from *baseline*. '''
    
    noiseline = baseline + alpha * np.sqrt(baseline + beta**2)
    
    return noiseline 



def _untilt(peak_idxs, y_hot, delta=5): 
    '''Adjust peak index positions to nearest true maxima.'''
    
    local_maxima = ssg.argrelextrema(y_hot, np.greater)[0] 
    
    new_peak_idxs = []

    for i in peak_idxs: 
    
        if i in list(local_maxima): 
            new_peak_idxs.append(i)

        else: 
            nearest = np.argmin((local_maxima - i)**2)
            i_new = local_maxima[nearest] 
            new_peak_idxs.append(i_new) 
            
    # check  
    for i in new_peak_idxs: 
        assert i in list(local_maxima), 'Peak adjustment error'
            
    return new_peak_idxs 


def _group_superimposed(hotmax_pixels): 
    '''Group superimposed `hotmax_pixels`. 
    
    Returns: hotmax_spots, hotmax_peak_idxs_list '''
    
    hotspots = hotmax_pixels[:,:2] 
    unique_hotspots = np.unique(hotspots, axis=0)
    
    # create ragged sorted list of indices 
    indices_list = [] 

    for u in unique_hotspots: 
        is_spot = np.all(hotspots == u, axis=1)
        indices = np.arange(len(hotspots))[is_spot].tolist()
        indices_list.append(indices)
        
    first_indices = [idxs[0] for idxs in indices_list]

    # sort spectrum indices according to lowest energy hotmax peak 
    sorting = np.argsort(first_indices)
    ragged_idxs = [indices_list[s] for s in sorting] 
    
    # create corresponding ragged list of peak channel indices 
    hotmax_peak_idxs_list = [list(hotmax_pixels[:,2][idxs]) for idxs in ragged_idxs] 
    
    # create corresponding list hotspots 
    hotmax_spots = np.array([hotmax_pixels[idxs[0]][0:2] for idxs in ragged_idxs])
    
    # DEPRECIATING BOOLEAN MASK: WILL USE append_list() INSTEAD  
    ## creat mask
    #n_hotspots = len(hotmax_spots)
    #shape = [n_hotspots, n_channels]
    #hotmax_mask = np.zeros(shape, dtype=bool)
    #
    #for i, idxs in enumerate(ragged_indices): 
    #    peak_idxs = hotmax_pixels[idxs][:, 2]
    #    hotmax_mask[i][peak_idxs] = True               
    #
    #return hotmax_spots, hotmax_mask 
    
    return hotmax_spots, hotmax_peak_idxs_list 
    



# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/10_storage.ipynb.

# %% auto 0
__all__ = ['DATASTACK_EXT', 'L', 'Layers', 'raw_to_datastack', 'tree', 'underscorify', 'append', 'append_list', 'repack',
           'max_and_sum_spectra', 'make_raw_preview', 'DataStack']

# %% ../notebooks/10_storage.ipynb 29
import maxrf4u

import numpy as np 
import dask 
import dask.array as da 
from dask.diagnostics import ProgressBar 
#import dask_ndfilters # obsolete 
from dask_image.ndfilters import gaussian_filter
import re 
import os 
import zarr 
from IPython.display import HTML 
import cv2
import matplotlib.pyplot as plt 
import scipy.signal as ssg 
import time 
import skimage as sk 

# %% ../notebooks/10_storage.ipynb 30
# CONSTANTS 
DATASTACK_EXT = '.datastack' 


# COMPUTION ORDER 

class Layers: 
    
    def __init__(self): 
        
        self.LAYERS = ['MAXRF_CUBE', 
                       'MAXRF_MAXSPECTRUM', 
                       'MAXRF_SUMSPECTRUM', 
                       'MAXRF_ENERGIES', 
                       'HOTMAX_PIXELS', 
                       'HOTMAX_SPECTRA', 
                       'HOTMAX_BASELINES', 
                       'HOTMAX_NOISELINES', 
                       'MAPS_IMVIS'] 
        
        for l in self.LAYERS: 
            setattr(self, l, l.lower())

L = Layers()




# functions 

def raw_to_datastack(raw_file, rpl_file, datastack_file=None, datapath=L.MAXRF_CUBE, verbose=True, 
                    flip_horizontal=False, flip_vertical=False): 
    '''Convert Bruker Macro XRF (.raw) data file *raw_filename* and (.rpl) shape file *rpl_filename*.  
    
    into a Zarr Zipstore datastack file (.datastack).''' 

    print('Please wait while preparing data conversion...')
    
    # generate datastack filename from .raw 
    if datastack_file is None: 
        datastack_file = re.sub('\.raw$', DATASTACK_EXT, raw_file) 
        
    # read data cube shape from .rpl file 
    with open(rpl_file, 'r') as fh: 
        lines = fh.readlines()

    # get rid of spaces and newline characters 
    keys_and_values = dict([re.sub(' |\n', '', l).split('\t') for l in lines]) 

    width = int(keys_and_values['width'])
    height = int(keys_and_values['height'])
    depth = int(keys_and_values['depth'])

    shape = (height, width, depth)
    
    # create numpy memory map with proper orientation 
    v_stride = 1
    h_stride = 1
    if flip_vertical: 
        v_stride = -1
    if flip_horizontal:
        h_stride = -1 
    
    print('Creating memory map...')
    raw_mm = np.memmap(raw_file, dtype='uint16', mode='r', shape=(height, width, depth))[::v_stride, ::h_stride] 

    # initializing dask array 
    arr = da.from_array(raw_mm) 
    arr = arr.astype(np.float32)
    

    # schedule spectral gaussian smoothing computation  
    smoothed = gaussian_filter(arr, (0, 0, 7)) 
    
    # create and open an empty zip file
    zs = zarr.ZipStore(datastack_file, mode='w') 
    
    if verbose: 
        print(f'Writing: {datastack_file}...')

    # compute and write maxrf data to zipstore 
    with ProgressBar(): 
        smoothed.to_zarr(zs, component=datapath) 
        
    zs.close()
    
    # compute sum and max spectra and append to zipstore 
    
    y_max, y_sum = max_and_sum_spectra(datastack_file, datapath=L.MAXRF_CUBE)
    
    append(y_max, L.MAXRF_MAXSPECTRUM, datastack_file)
    append(y_sum, L.MAXRF_SUMSPECTRUM, datastack_file)
    
    

def tree(datastack_file, show_arrays=False): 
    '''Prints content tree of *datastack_file* '''

    with zarr.ZipStore(datastack_file, mode='r') as zs: 
        root = zarr.group(store=zs) 
        tree = root.tree(expand=True).__repr__()
        print(f'{datastack_file}:\n\n{tree}')  
        
        if show_arrays:        
            datasets = sorted(root)
            arrays_html = ''

            for ds in datasets: 
                arr = da.from_array(root[ds])
                html = arr._repr_html_()
                arrays_html = f'{arrays_html}- Dataset: <h style="color:brown">{ds}</h>{html}' 
   
            return HTML(arrays_html)

def underscorify(datapath, datapath_list, extra_underscore=True): 
    '''Append extra underscore if *datapath* exists to prevent overwriting. 
    
    If *extra_underscore=False* return (latest) datapath with most underscores'''
    
    if datapath in datapath_list: 
        r = re.compile(f'{datapath}_*$')
        datapath = sorted(filter(r.match, datapath_list))[-1]
        
        if extra_underscore: 
            datapath = datapath + '_'
        
    return datapath 


def append(arr, datapath, datastack_file): 
    '''Add numpy or dask array *arr* to *datastack_file* in folder *datapath*.'''
    
    if not isinstance(arr, dask.array.Array):  
        arr = da.from_array(arr) 
            
    with zarr.ZipStore(datastack_file, mode='a') as zs: 
        root = zarr.group(store=zs)
        
        # append underscores to make unique if datapath exists 
        datapath_list = sorted(root) 
        datapath = underscorify(datapath, datapath_list)
        
        # write      
        arr.to_zarr(zs, component=datapath)

        
def append_list(ragged_list, datapath, datastack_file, nan=-9999): 
    '''Wrapper around append() to store iregular (ragged) lists of lists as regular padded arrays.  
    
    Currently only working for two dimensional lists of integers. Padding is done with nan=-9999. 
    ''' 
    
    padded_array = _straighten(ragged_list, nan=nan) 
        
    append(padded_array, datapath, datastack_file) 
    

def repack(datastack_file, select='all', overwrite=True, verbose=False): 
    '''Repack *datastack_file* by deleting and renaming all but latest datasets. 
    
    Automatic selection of latest datasets can be overriden be providing list of *select* datasets''' 
    
    if verbose: 
        tree(datastack_file)
    
    # open existing zipstore  
    zs = zarr.ZipStore(datastack_file, mode='r') 
    root = zarr.group(store=zs)
    datapath_list = sorted(root)  
    
    # select newest version (most underscores) for all datasets
    if select == 'all': 
        selected = sorted(set([underscorify(dp, datapath_list, extra_underscore=False) for dp in datapath_list])) 
    # select newest version (most underscores) for datasets in select
    else: 
        selected = sorted(set([underscorify(dp, datapath_list, extra_underscore=False) for dp in select]))       
    
    # remove underscores 
    renamed = [re.sub('_*$', '', s) for s in selected] 
    
    # create and open new empty zipstore 
    datastack_file_new = datastack_file + '_temp'
    zs_new = zarr.ZipStore(datastack_file_new, mode='w') 
    
    # copy selected datasets into new zipstore 
    with ProgressBar(): 
        for src, dst in zip(selected, renamed): 
            print(f'Repacking dataset: \'{src}\'') 
            arr = da.from_array(root[src])
            arr.to_zarr(zs_new, component=dst)
    
    zs.close()
    zs_new.close()
    
    # finally overwrite old with new  
    if overwrite: 
        os.replace(datastack_file_new, datastack_file)
    
    if verbose:
        print()
        tree(datastack_file)
        

def max_and_sum_spectra(datastack_file, datapath=L.MAXRF_CUBE): 
    '''Compute sum spectrum and max spectrum for 'maxrf' dataset in *datastack_file*. 
    
    Returns: *y_sum*, *y_max*'''
    
    # open existing zipstore  
    zs = zarr.ZipStore(datastack_file, mode='r') 
    root = zarr.group(store=zs)
    
    # initialize dask array 
    arr = da.from_array(root[datapath])
        
    # flatten (better avoid)
    h, w, d = arr.shape 
    #flat_shape = h * w, d
    #with dask.config.set(**{'array.slicing.split_large_chunks': True}):
    #    arr_flat = arr.reshape(flat_shape) #, limit='128 MiB') 
    #
    ## schedule computations 
    #sum_spectrum = arr_flat.sum(axis=0)
    #max_spectrum = arr_flat.max(axis=0)
    sum_spectrum = arr.sum(axis=(0, 1))
    max_spectrum = arr.max(axis=(0, 1))
    
    # compute 
    print('Computing max spectrum...')
    with ProgressBar():
        y_max = max_spectrum.compute() 
    print('Computing sum spectrum...')
    with ProgressBar(): 
        y_sum = sum_spectrum.compute() / (h * w)
        
    zs.close()
     
    return y_max, y_sum 


def make_raw_preview(raw_file, rpl_file, show=False): 
    '''
    Create preview image of raw file to inspect scan orientation. 
    '''

    # read data cube shape from .rpl file 
    with open(rpl_file, 'r') as fh: 
        lines = fh.readlines()
    
    # get rid of spaces and newline characters 
    keys_and_values = dict([re.sub(' |\n', '', l).split('\t') for l in lines]) 
    
    width = int(keys_and_values['width'])
    height = int(keys_and_values['height'])
    depth = int(keys_and_values['depth'])
    
    shape = (height, width, depth)
    
    # create numpy memory map 
    raw_mm = np.memmap(raw_file, dtype='uint16', mode='r', shape=(height, width, depth)) #[::-1, ::-1] 
    
    # create max-spectrum 
    raw_flat = raw_mm.reshape([-1, depth])
    raw_max = np.max(raw_flat, axis=0)
    
    # locate highest peak 
    max_peak_idx = np.argmax(raw_max)
    
    # integrate peak slice 
    
    max_peak_map = np.average(raw_mm[:,:,max_peak_idx-10:max_peak_idx+10], axis=2)
    raw_preview = sk.exposure.equalize_hist(max_peak_map) 
    
    preview_fname = raw_file + '_preview.png'
    print(f'Saving: {preview_fname}...')
    plt.imsave(preview_fname, raw_preview) 

    if show: 
        fig, ax = plt.subplots()
        ax.imshow(raw_preview)
        ax.set_title(preview_fname)



        
class DataStack: 
        
    def __init__(self, datastack_file, mode='r', verbose=False, show_arrays=True): 
        '''Initialize DataStack object from *datastack_file*.''' 
        
        # default computation layers ordering as attributes  
        
        self.LAYERS = L.LAYERS 
        
        for l in L.LAYERS: 
            setattr(self, l, l.lower())
            
        # read datasets from file  
        
        self.mode = mode 
        self.datastack_file = datastack_file 
        
        self.update_attrs()
            
        # print tree 
        if verbose: 
            tree(self.datastack_file, show_arrays=show_arrays) 
            
    def update_attrs(self): 
        
        # populate store attributes 
        self.store = zarr.ZipStore(self.datastack_file, mode=self.mode) 
        self.root = zarr.group(store=self.store) 
        
        # generic exposure to dask arrays 
        self.datapath_list = sorted(self.root) 
        self.datasets = {dp: da.from_array(self.root[dp]) for dp in self.datapath_list}
        
        # attributify dask arrays 
        # useful for code development, perhaps confusing for users 
        # might turn off this feature later 
        for dp, ds in self.datasets.items(): 
            setattr(self, dp, ds) 
        
            
    def latest(self, datapath): 
        '''Return latest version of datapath. '''
        
        datapath = underscorify(datapath, self.datapath_list, extra_underscore=False)
        
        return datapath 
        
            
    def read(self, datapath, latest=True, compute=True):
        '''Read latest version of dataset for *datapath*
        
        Returns numpy array if dataset exists. Otherwise exits. '''
        
        if datapath in self.datapath_list: 
            if latest: 
                datapath = self.latest(datapath)     
            dataset = self.datasets[datapath] 
            if compute: 
                dataset = dataset.compute()
                
        # no dataset in file        
        else: 
            dataset = None 
            
            self.tree()
            assert False, f'Dataset not found: {datapath}'
    
        return dataset
    
    
    def read_list(self, datapath, latest=True, nan=-9999): 
        '''Thin wrapper for reading padded arrays (ragged lists). 

        Returns ragged list if dataset exists. Current implementation only for 
        two-dimensional (ragged) list of lists. ''' 

        # step 1: read (padded array 
        padded_array = self.read(datapath, latest=latest, compute=True)

        # step 2: convert to ragged list by removing nan values 
        ragged_list = _unstraighten(padded_array, nan=nan)

        return ragged_list 


    
    def tree(self, show_arrays=False): 
        '''Prints content tree of datastack.'''
        
        tree(self.datastack_file, show_arrays=show_arrays)
            

            
    def close(self): 
        '''Close file handle'''
         
        self.store.close()
        self.mode = 'closed' 
        
        print(f'Closed: {self.datastack_file}')             

def _straighten(ragged_list, nan=-9999): 
    '''Utility function to straighten a `ragged_list` of integers indices into a regular (padded) array. 
    
    
    Creates a two dimensional numpy array with empty values padded with nan=-9999. 
    
    Returns: padded_array 
    '''
    
    # determine shape 
    ncols = max([len(idxs) for idxs in ragged_list])
    nrows = len(ragged_list) 
    
    # initialize 
    padded_array = np.ones([nrows, ncols], dtype=int)
    padded_array[:,:] = nan 
    
    # fill 
    
    for i, indices in enumerate(ragged_list): 
        for j, idx in enumerate(indices): 
            padded_array[i, j] = idx 
        
    return padded_array 
    

def _unstraighten(padded_array, nan=-9999):
    '''Convert a numpy `padded_array` of integers filled out with nan's into a ragged list.
    
    
    Returns: a ragged list of lists 
    '''

    ragged_list = []
    
    for row in padded_array: 
        row_list = list(row[row!=nan]) # remove nan's from list  
        ragged_list.append(row_list)

    return ragged_list
    
   



"""Matching peak patterns"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/50_peak-pattern-puzzle.ipynb.

# %% auto 0
__all__ = ['EOI', 'Peak_Pattern_Atlas', 'plot_patterns', 'get_eoi_patterns', 'colorize', 'get_instrument_pattern', 'update_eoi']

# %% ../notebooks/50_peak-pattern-puzzle.ipynb 56
import maxrf4u  

from . import HotmaxAtlas 
from . import DataStack
from .peakmaps import _add_hotlines_ticklabels 

from .xphysics import get_element_spectrum 


import re 
import matplotlib.pyplot as plt 
import matplotlib.cm as cm 
import mendeleev
import moseley as mos
import numpy as np 

# %% ../notebooks/50_peak-pattern-puzzle.ipynb 57
class Peak_Pattern_Atlas():
    
    def __init__(self, elements_of_interest=None, tube_keV=40, datastack_file=None, verbose=True): 
        '''Compute atlas of peak patterns for all `elements_of_interest`. '''

        if verbose: 
            print('Please wait while initializing peak pattern atlas...')
        if elements_of_interest == None:             
            self.elements_of_interest = ['#H', '#He', '#Li', '#Be', '#B', '#C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg', 'Al', 
                                         'Si', 'P', 'S', 'Cl', '#Ar', 'K', 'Ca', '#Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 
                                         'Co', 'Ni', 'Cu', 'Zn', '#Ga', '#Ge', 'As', '#Se', 'Br', '#Kr', '#Rb', 'Sr', 
                                         '#Y', '#Zr', '#Nb', '#Mo', '#Tc', '#Ru', 'Rh', '#Pd', 'Ag', 'Cd', '#In', 'Sn', 
                                         '#Sb', '#Te', 'I', '#Xe', '#Cs', 'Ba', '#La', '#Hf', '#Ta', '#W', '#Re', '#Os', 
                                         '#Ir', '#Pt', 'Au', 'Hg', '#Tl', 'Pb', '#Bi', '#Po', '#At', '#Rn', '#Fr', '#Ra', 
                                         '#Ac', '#Rf', '#Db', '#Sg', '#Bh', '#Hs', '#Mt', '#Ds', '#Rg', '#Cn', '#Nh', 
                                         '#Fl', '#Mc', '#Lv', '#Ts', '#Og'] 
        else: 
            self.elements_of_interest = elements_of_interest
        self.tube_keV = tube_keV    
        self.eoi_ptrns = get_eoi_patterns(self.elements_of_interest, tube_keV=tube_keV) 
        
        # find right most marker to set plot limit based on pattern atlas 
        self.xmax = max([max(ptrn['peaks_xy'][:, 0]) for ptrn in self.eoi_ptrns])
        
        if datastack_file != None: 
            self.datastack_file = datastack_file 
            self.instrum_ptrn = get_instrument_pattern(datastack_file)
            self.hma = HotmaxAtlas(datastack_file)
            self.contains_dict = {}
            
        else:
            self.instrum_ptrn = None 

    def update_eoi(self, add_elems): 
        '''Update elements of interest and corresponding eoi_ptrns . '''

        self.elements_of_interest = update_eoi(self.elements_of_interest, add_elems)
        self.eoi_ptrns = get_eoi_patterns(self.elements_of_interest, tube_keV=self.tube_keV) 
        self.xmax = max([max(ptrn['peaks_xy'][:, 0]) for ptrn in self.eoi_ptrns])
        
        
                         
    def plot_patterns(self, select_elems=None, ax=None): 
        '''
        Plot peak patterns for list of elements `plot_elems`.
        '''

        self._ax = plot_patterns(self.eoi_ptrns, select_elems=select_elems, instrum_ptrn=self.instrum_ptrn, ax=ax)
        ymin, ymax = self._ax.get_ylim()
        
        # add max spectrum vlines in pattern plot 
        if True: #isinstance(self.hma, HotmaxAtlas):
            self.lines_x = self.hma.x_keVs[self.hma.hotmax_peak_idxs_flat]
            self._ax.vlines(self.lines_x, ymin=ymin, ymax=ymax, color='r', alpha=0.2, zorder=-10) 
            
        self._ax.set_xlim([-1, self.xmax + 2])   



    def plot_puzzle(self, spectrum_n, select_elems=None): 
        
        # store selected elements for spectrum n
        self.contains_dict[spectrum_n] = select_elems

        fig, axs = plt.subplots(nrows=2, figsize=[10, 8], sharex=True, height_ratios=[4, 1])
        
        ax0, ax1 = axs.flatten()
        
        self.plot_patterns(select_elems=select_elems, ax=ax0) 
        ax, labels = self.hma.plot_spectrum(spectrum_n, ax=ax1, xlim_padding=None)
        ax1.set_xlabel('energy [keV]')
        plt.subplots_adjust(hspace=0.02)

        plt.tight_layout()

        


def plot_patterns(eoi_ptrns, select_elems=None, instrum_ptrn=None, ax=None): 
    '''Wrapper function to plot overview of element patterns `elem_ptrns` and instrument patterns in axes `ax`. 
   
    Returns: `ax` 
    ''' 

    # select all elements if None
    if select_elems == None: 
        select_elems = []
    #else: 
    #
    #    # underscore implies grey color 
    #    greyed_elems = [re.sub('_', '', elem) for elem in plot_elems if '_' in elem] #  
    #     
    #    # remove underscores and create selected elements pattern list to plot 
    #    selected_elems = [re.sub('_', '', elem) for elem in plot_elems] # unsorted! 
    #    ptrns = [ptrn for ptrn in eoi_ptrns if ptrn['elem'] in selected_elems] # sorted 

    elements = [ptrn['elem'] for ptrn in eoi_ptrns]
    names = [ptrn['name'] for ptrn in eoi_ptrns]
    Z = [ptrn['atomic_number'] for ptrn in eoi_ptrns]
    
    
    ytick_labels = [f'{n}({z})]' for n, z in zip(names, Z)] 
    
    n_ptrns = len(eoi_ptrns)
    n_ticks = n_ptrns 
    offset = 0 

    
    # create extra vertical space in plot 
    if instrum_ptrn != None: 
        
        # add space for instrument pattern 
        n_ticks += 1 
        offset = 1 
        ytick_labels = ['INSTRUMENT'] + ytick_labels 
        
    # now create plot    
    if ax is None:  
        fig_width = 9
        fig_height = 0.15 * n_ticks + 0.5 
        fig, ax = plt.subplots(figsize=[fig_width, fig_height])  
        ax.set_xlabel('energy (keV)')
            
    # plot instrument pattern X X X 
    if instrum_ptrn != None: 
        instr_peaks = instrum_ptrn['instrument_peaks']
        instr_zeros = np.zeros_like(instr_peaks) 
        ax.scatter(instr_peaks, instr_zeros, marker='x', color='r')   
    
    # plot element patterns     
    for i, ptrn in enumerate(eoi_ptrns):        
        element = ptrn['elem']
        x, _ = ptrn['peaks_xy'].T
        y = np.zeros_like(x) + i + offset
        color = ptrn['color']
        if element in select_elems: 
            facecolor = color 
            edgecolor = 'k'
            linecolor = color
            linestyle = '-'
        else: 
            facecolor = 'w'
            edgecolor = 'grey'
            linecolor = 'grey'
            linestyle = ':'
            

        alpha_escape_keV = ptrn['alpha_escape_keV']
         
        left_x = min(x)
        right_x = max(x)
        
        ax.scatter(x, y, s=15, facecolor=facecolor, edgecolor=edgecolor)
        ax.scatter(x[0], y[0], marker='s', facecolor=facecolor, edgecolor=edgecolor)
        ax.plot([left_x, right_x], [y[0], y[0]], color=edgecolor, linestyle=linestyle, zorder=-10)
        ax.scatter(alpha_escape_keV, y[0], marker='|', color=edgecolor) 
        ax.annotate(element, [right_x, y[0]], xytext=[5, -1], fontsize=8, color='k', 
                    textcoords='offset points', ha='left', va='center') 
                  
    ax.set_yticks(range(n_ticks))        
    ax.set_yticklabels(ytick_labels, fontsize=8) 
        
    ax.set_ylim([-1, n_ticks])
    
    return ax 




#|export 

# all elements 
EOI = ['#H', '#He', '#Li', '#Be', '#B', '#C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg', 'Al', 
       'Si', 'P', 'S', 'Cl', '#Ar', 'K', 'Ca', '#Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 
       'Co', 'Ni', 'Cu', 'Zn', '#Ga', '#Ge', 'As', '#Se', 'Br', '#Kr', '#Rb', 'Sr', 
       '#Y', '#Zr', '#Nb', '#Mo', '#Tc', '#Ru', 'Rh', '#Pd', 'Ag', 'Cd', '#In', 'Sn', 
       '#Sb', '#Te', 'I', '#Xe', '#Cs', 'Ba', '#La', '#Hf', '#Ta', '#W', '#Re', '#Os', 
       '#Ir', '#Pt', 'Au', 'Hg', '#Tl', 'Pb', '#Bi', '#Po', '#At', '#Rn', '#Fr', '#Ra', 
       '#Ac', '#Rf', '#Db', '#Sg', '#Bh', '#Hs', '#Mt', '#Ds', '#Rg', '#Cn', '#Nh', 
       '#Fl', '#Mc', '#Lv', '#Ts', '#Og']


def get_eoi_patterns(elements_of_interest, tube_keV=40): 
    '''
    Create a list of xrf spectrum pattern  dictionaries for all `elements_of_interest`. 

    Hashed (#) elements are excluded from the list.  
    Returns `eoi_ptrns` dictionary list. 
    '''
     
    ptrn_dict_list = [] 

    # remove hashed (#) elements 
    elements = [e for e in elements_of_interest if not '#' in e] 
    if elements is None: 
        elements = EOI 
    
    for elem in elements: 
        
        # xrf physics 
        spectrum = mos.XFluo(elem, tube_keV=tube_keV, min_prom=0.01)
        peak_order = np.argsort(spectrum.peak_intensities)[::-1]
        alpha_keV = spectrum.peak_energies[peak_order][0]
        peaks_x = spectrum.peak_energies[peak_order]
        peaks_y = spectrum.peak_intensities[peak_order]
        peaks_xy = np.c_[peaks_x, peaks_y] 
        alpha_escape_keV = alpha_keV - 1.74  # Silicon detector escape energy shift 
         
        # element       
        name = mendeleev.element(elem).name
        atomic_number = mendeleev.element(elem).atomic_number

        # standard color map 
        color = colorize(elem) 
        
        ptrn_dict = {'elem': elem,  
                     'atomic_number': atomic_number, 
                     'name': name, 
                     'peaks_xy': peaks_xy, 
                     'alpha_escape_keV': alpha_escape_keV, 
                     'color': color}
        
        ptrn_dict_list.append(ptrn_dict) 
        
    # sort according to alpha energy     
    alpha_keVs = [p['peaks_xy'][0, 0] for p in ptrn_dict_list]        
    indices = np.argsort(alpha_keVs) 
    
    eoi_ptrns = [ptrn_dict_list[i] for i in indices] 
        
    return eoi_ptrns


def colorize(elem): 
    '''Pick fixed color from nice color map for elements of interest. '''
    
    # select elements of interest 
    all_elements = ['#H', '#He', '#Li', '#Be', '#B', '#C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg', 'Al', 
                    'Si', 'P', 'S', 'Cl', '#Ar', 'K', 'Ca', '#Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 
                    'Co', 'Ni', 'Cu', 'Zn', '#Ga', '#Ge', 'As', '#Se', 'Br', '#Kr', '#Rb', 'Sr', 
                    '#Y', '#Zr', '#Nb', '#Mo', '#Tc', '#Ru', 'Rh', '#Pd', 'Ag', 'Cd', '#In', 'Sn', 
                    '#Sb', '#Te', 'I', '#Xe', '#Cs', 'Ba', '#La', '#Hf', '#Ta', '#W', '#Re', '#Os', 
                    '#Ir', '#Pt', 'Au', 'Hg', '#Tl', 'Pb', '#Bi', '#Po', '#At', '#Rn', '#Fr', '#Ra', 
                    '#Ac', '#Rf', '#Db', '#Sg', '#Bh', '#Hs', '#Mt', '#Ds', '#Rg', '#Cn', '#Nh', 
                    '#Fl', '#Mc', '#Lv', '#Ts', '#Og']

    eoi = [e for e in all_elements if not '#' in e] 
        
    # tab20x2 color map 
    tab20 = cm.tab20(np.arange(20))[:,0:3]
    colors = np.r_[tab20[::-1], tab20[::-1]**0.6]
    
    # this code logic is pretty ad hoc (needs improvement)
    # custom colors 
    special_colors = {'Pb': (0.4, 0.4, 0.4), 
                      'Cu': (0.1, 0.9, 0.3), 
                      'Fe': (0.7, 0.5, 0.1), 
                      'S': (1.0, 0.9, 0.1),
                      'Au': (1, 0.9, 0.1), 
                      'Br': (0.4, 0.3, 0), 
                      'Rh': (0.3, 0.1, 0.1)} 
    
    for e in special_colors.keys(): 
        if e in eoi: 
            colors[eoi.index(e)] = special_colors[e]
        
    if elem in eoi:  
            color = colors[eoi.index(elem)]  
            
    else: 
        color = [0.7, 0.7, 0.7]
    
    return color 
       



def get_instrument_pattern(datastack_file):
    '''Generate instrument peak pattern. 
    
    Pattern dictionary contains strongest Rhodium anode emission peaks,  
    their corresponding Compton shifted peaks, and a sensor peak 
    
    Returns: instrument_pattern_dict 
    
    '''
    
    ds = maxrf4u.DataStack(datastack_file) 
    
    x_keVs = ds.read('maxrf_energies')
    keV0 = ds.read('compton_peak_energy')[0]
    keV1 = maxrf4u.RHODIUM_Ka 
    theta = maxrf4u.detector_angle(keV0, keV1) 
    
    sensor_peak_idx = ds.read_list('hotmax_peak_idxs_list')[0][0]
    sensor_peak_keV = x_keVs[sensor_peak_idx]
    

    anode_emission = mos.XFluo('Rh', tube_keV=30, min_prom=0.1) # only largest K peaks  
    keV_in = anode_emission.peak_energies
    keV_out = maxrf4u.compton_shift(keV_in, theta)

    instrument_peaks = np.r_[sensor_peak_keV, keV_in, keV_out]
    instrument_peaks = np.sort(instrument_peaks) 
    
    instrument_pattern_dict = {'name': 'INSTRUMENT', 
                               'instrument_peaks': instrument_peaks}
    
    return instrument_pattern_dict
    


def update_eoi(elements_of_interest, add_elems):
    '''
    Update `elements_of_interest` list for elements list `add_elems`. 

    For example: add_elems=['#Al', 'Ga]
    '''
    
    for elem in add_elems: 
        if not '#' in elem: 
            hashed_elem = f'#{elem}'
            if hashed_elem in elements_of_interest: 
                idx = elements_of_interest.index(hashed_elem)
                elements_of_interest[idx] = elem 
        elif '#' in elem: 
            unhashed_elem = re.sub('#', '', elem)
            if unhashed_elem in elements_of_interest:
                idx = elements_of_interest.index(unhashed_elem)
                elements_of_interest[idx] = elem

    return elements_of_interest



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/maxrf4u',
                'doc_host': 'https://fligt.github.io',
                'git_url': 'https://github.com/fligt/maxrf4u',
                'lib_path': 'maxrf4u'},
  'syms': { 'maxrf4u.absorption': {},
            'maxrf4u.calibration': { 'maxrf4u.calibration.calibrate': ('calibration.html#calibrate', 'maxrf4u/calibration.py'),
                                     'maxrf4u.calibration.compton_shift': ('calibration.html#compton_shift', 'maxrf4u/calibration.py'),
                                     'maxrf4u.calibration.detector_angle': ('calibration.html#detector_angle', 'maxrf4u/calibration.py'),
                                     'maxrf4u.calibration.find_instrument_peaks': ( 'calibration.html#find_instrument_peaks',
                                                                                    'maxrf4u/calibration.py')},
            'maxrf4u.fingerprints': {},
            'maxrf4u.hotmax': { 'maxrf4u.hotmax.HotmaxAtlas': ('hotmax.html#hotmaxatlas', 'maxrf4u/hotmax.py'),
                                'maxrf4u.hotmax.HotmaxAtlas.__init__': ('hotmax.html#hotmaxatlas.__init__', 'maxrf4u/hotmax.py'),
                                'maxrf4u.hotmax.HotmaxAtlas.plot_spectra': ('hotmax.html#hotmaxatlas.plot_spectra', 'maxrf4u/hotmax.py'),
                                'maxrf4u.hotmax.HotmaxAtlas.plot_spectrum': ('hotmax.html#hotmaxatlas.plot_spectrum', 'maxrf4u/hotmax.py'),
                                'maxrf4u.hotmax._baseline': ('hotmax.html#_baseline', 'maxrf4u/hotmax.py'),
                                'maxrf4u.hotmax._group_superimposed': ('hotmax.html#_group_superimposed', 'maxrf4u/hotmax.py'),
                                'maxrf4u.hotmax._noiseline': ('hotmax.html#_noiseline', 'maxrf4u/hotmax.py'),
                                'maxrf4u.hotmax._untilt': ('hotmax.html#_untilt', 'maxrf4u/hotmax.py'),
                                'maxrf4u.hotmax.compute_hotmax_noise': ('hotmax.html#compute_hotmax_noise', 'maxrf4u/hotmax.py'),
                                'maxrf4u.hotmax.compute_hotmax_spectra': ('hotmax.html#compute_hotmax_spectra', 'maxrf4u/hotmax.py'),
                                'maxrf4u.hotmax.compute_subpeaks': ('hotmax.html#compute_subpeaks', 'maxrf4u/hotmax.py')},
            'maxrf4u.interactive_visualization': { 'maxrf4u.interactive_visualization.UploadDir': ( 'interactive-plotting.html#uploaddir',
                                                                                                    'maxrf4u/interactive_visualization.py'),
                                                   'maxrf4u.interactive_visualization.UploadDir.__init__': ( 'interactive-plotting.html#uploaddir.__init__',
                                                                                                             'maxrf4u/interactive_visualization.py'),
                                                   'maxrf4u.interactive_visualization.UploadDir.export_interactive_html': ( 'interactive-plotting.html#uploaddir.export_interactive_html',
                                                                                                                            'maxrf4u/interactive_visualization.py'),
                                                   'maxrf4u.interactive_visualization.UploadDir.imsave': ( 'interactive-plotting.html#uploaddir.imsave',
                                                                                                           'maxrf4u/interactive_visualization.py'),
                                                   'maxrf4u.interactive_visualization.UploadDir.listdir': ( 'interactive-plotting.html#uploaddir.listdir',
                                                                                                            'maxrf4u/interactive_visualization.py'),
                                                   'maxrf4u.interactive_visualization.UploadDir.listurls': ( 'interactive-plotting.html#uploaddir.listurls',
                                                                                                             'maxrf4u/interactive_visualization.py'),
                                                   'maxrf4u.interactive_visualization.make_filenames': ( 'interactive-plotting.html#make_filenames',
                                                                                                         'maxrf4u/interactive_visualization.py'),
                                                   'maxrf4u.interactive_visualization.make_gridbox_widget': ( 'interactive-plotting.html#make_gridbox_widget',
                                                                                                              'maxrf4u/interactive_visualization.py')},
            'maxrf4u.iplotting': {},
            'maxrf4u.peak_pattern_puzzle_solver': {},
            'maxrf4u.peakmaps': { 'maxrf4u.peakmaps._add_hotlines_ticklabels': ( 'peakmaps.html#_add_hotlines_ticklabels',
                                                                                 'maxrf4u/peakmaps.py'),
                                  'maxrf4u.peakmaps._fit_gaussian': ('peakmaps.html#_fit_gaussian', 'maxrf4u/peakmaps.py'),
                                  'maxrf4u.peakmaps._gaussian': ('peakmaps.html#_gaussian', 'maxrf4u/peakmaps.py'),
                                  'maxrf4u.peakmaps._get_spectral_slices': ('peakmaps.html#_get_spectral_slices', 'maxrf4u/peakmaps.py'),
                                  'maxrf4u.peakmaps.compute_nmf_element_maps': ( 'peakmaps.html#compute_nmf_element_maps',
                                                                                 'maxrf4u/peakmaps.py'),
                                  'maxrf4u.peakmaps.compute_nmf_peakmaps': ('peakmaps.html#compute_nmf_peakmaps', 'maxrf4u/peakmaps.py'),
                                  'maxrf4u.peakmaps.fit_spectrum_peaks': ('peakmaps.html#fit_spectrum_peaks', 'maxrf4u/peakmaps.py'),
                                  'maxrf4u.peakmaps.get_continuum': ('peakmaps.html#get_continuum', 'maxrf4u/peakmaps.py'),
                                  'maxrf4u.peakmaps.get_gaussians': ('peakmaps.html#get_gaussians', 'maxrf4u/peakmaps.py'),
                                  'maxrf4u.peakmaps.multi_plot': ('peakmaps.html#multi_plot', 'maxrf4u/peakmaps.py')},
            'maxrf4u.peakpuzzle': { 'maxrf4u.peakpuzzle.Peak_Pattern_Atlas': ( 'peak-pattern-puzzle.html#peak_pattern_atlas',
                                                                               'maxrf4u/peakpuzzle.py'),
                                    'maxrf4u.peakpuzzle.Peak_Pattern_Atlas.__init__': ( 'peak-pattern-puzzle.html#peak_pattern_atlas.__init__',
                                                                                        'maxrf4u/peakpuzzle.py'),
                                    'maxrf4u.peakpuzzle.Peak_Pattern_Atlas.plot_patterns': ( 'peak-pattern-puzzle.html#peak_pattern_atlas.plot_patterns',
                                                                                             'maxrf4u/peakpuzzle.py'),
                                    'maxrf4u.peakpuzzle.Peak_Pattern_Atlas.plot_puzzle': ( 'peak-pattern-puzzle.html#peak_pattern_atlas.plot_puzzle',
                                                                                           'maxrf4u/peakpuzzle.py'),
                                    'maxrf4u.peakpuzzle.Peak_Pattern_Atlas.update_eoi': ( 'peak-pattern-puzzle.html#peak_pattern_atlas.update_eoi',
                                                                                          'maxrf4u/peakpuzzle.py'),
                                    'maxrf4u.peakpuzzle.colorize': ('peak-pattern-puzzle.html#colorize', 'maxrf4u/peakpuzzle.py'),
                                    'maxrf4u.peakpuzzle.get_eoi_patterns': ( 'peak-pattern-puzzle.html#get_eoi_patterns',
                                                                             'maxrf4u/peakpuzzle.py'),
                                    'maxrf4u.peakpuzzle.get_instrument_pattern': ( 'peak-pattern-puzzle.html#get_instrument_pattern',
                                                                                   'maxrf4u/peakpuzzle.py'),
                                    'maxrf4u.peakpuzzle.plot_patterns': ('peak-pattern-puzzle.html#plot_patterns', 'maxrf4u/peakpuzzle.py'),
                                    'maxrf4u.peakpuzzle.update_eoi': ('peak-pattern-puzzle.html#update_eoi', 'maxrf4u/peakpuzzle.py')},
            'maxrf4u.speckles': {},
            'maxrf4u.storage': { 'maxrf4u.storage.DataStack': ('storage.html#datastack', 'maxrf4u/storage.py'),
                                 'maxrf4u.storage.DataStack.__init__': ('storage.html#datastack.__init__', 'maxrf4u/storage.py'),
                                 'maxrf4u.storage.DataStack.close': ('storage.html#datastack.close', 'maxrf4u/storage.py'),
                                 'maxrf4u.storage.DataStack.latest': ('storage.html#datastack.latest', 'maxrf4u/storage.py'),
                                 'maxrf4u.storage.DataStack.read': ('storage.html#datastack.read', 'maxrf4u/storage.py'),
                                 'maxrf4u.storage.DataStack.read_list': ('storage.html#datastack.read_list', 'maxrf4u/storage.py'),
                                 'maxrf4u.storage.DataStack.tree': ('storage.html#datastack.tree', 'maxrf4u/storage.py'),
                                 'maxrf4u.storage.DataStack.update_attrs': ('storage.html#datastack.update_attrs', 'maxrf4u/storage.py'),
                                 'maxrf4u.storage.Layers': ('storage.html#layers', 'maxrf4u/storage.py'),
                                 'maxrf4u.storage.Layers.__init__': ('storage.html#layers.__init__', 'maxrf4u/storage.py'),
                                 'maxrf4u.storage._straighten': ('storage.html#_straighten', 'maxrf4u/storage.py'),
                                 'maxrf4u.storage._unstraighten': ('storage.html#_unstraighten', 'maxrf4u/storage.py'),
                                 'maxrf4u.storage.append': ('storage.html#append', 'maxrf4u/storage.py'),
                                 'maxrf4u.storage.append_list': ('storage.html#append_list', 'maxrf4u/storage.py'),
                                 'maxrf4u.storage.make_raw_preview': ('storage.html#make_raw_preview', 'maxrf4u/storage.py'),
                                 'maxrf4u.storage.max_and_sum_spectra': ('storage.html#max_and_sum_spectra', 'maxrf4u/storage.py'),
                                 'maxrf4u.storage.raw_to_datastack': ('storage.html#raw_to_datastack', 'maxrf4u/storage.py'),
                                 'maxrf4u.storage.repack': ('storage.html#repack', 'maxrf4u/storage.py'),
                                 'maxrf4u.storage.tree': ('storage.html#tree', 'maxrf4u/storage.py'),
                                 'maxrf4u.storage.underscorify': ('storage.html#underscorify', 'maxrf4u/storage.py')},
            'maxrf4u.warp': { 'maxrf4u.warp.ImageRegistrationHelper': ( 'spatial-verus-spectral.html#imageregistrationhelper',
                                                                        'maxrf4u/warp.py'),
                              'maxrf4u.warp.ImageRegistrationHelper.__init__': ( 'spatial-verus-spectral.html#imageregistrationhelper.__init__',
                                                                                 'maxrf4u/warp.py'),
                              'maxrf4u.warp.ImageRegistrationHelper.get_marker_coordinates': ( 'spatial-verus-spectral.html#imageregistrationhelper.get_marker_coordinates',
                                                                                               'maxrf4u/warp.py'),
                              'maxrf4u.warp.ImageRegistrationHelper.show': ( 'spatial-verus-spectral.html#imageregistrationhelper.show',
                                                                             'maxrf4u/warp.py'),
                              'maxrf4u.warp._img_to_base64_url': ('spatial-verus-spectral.html#_img_to_base64_url', 'maxrf4u/warp.py'),
                              'maxrf4u.warp.latlon_to_xy': ('spatial-verus-spectral.html#latlon_to_xy', 'maxrf4u/warp.py'),
                              'maxrf4u.warp.warp': ('spatial-verus-spectral.html#warp', 'maxrf4u/warp.py'),
                              'maxrf4u.warp.xy_to_latlon': ('spatial-verus-spectral.html#xy_to_latlon', 'maxrf4u/warp.py')},
            'maxrf4u.xphysics': { 'maxrf4u.xphysics.ElementLines': ('xphysics.html#elementlines', 'maxrf4u/xphysics.py'),
                                  'maxrf4u.xphysics.ElementLines.__init__': ('xphysics.html#elementlines.__init__', 'maxrf4u/xphysics.py'),
                                  'maxrf4u.xphysics.gaussian_convolve': ('xphysics.html#gaussian_convolve', 'maxrf4u/xphysics.py'),
                                  'maxrf4u.xphysics.get_element_spectra': ('xphysics.html#get_element_spectra', 'maxrf4u/xphysics.py'),
                                  'maxrf4u.xphysics.get_element_spectrum': ('xphysics.html#get_element_spectrum', 'maxrf4u/xphysics.py')}}}
